# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk

from solfege import abstract
from solfege import cfg
from solfege import gu
from solfege import lessonfile
from solfege import mpd
from solfege import soundcard
from solfege import utils
from solfege.mpd import Rat

import solfege


class Teacher(abstract.Teacher):
    """
    The Teacher and Gui for dictation abuses q_status a little.
    QSTATUS_NEW mean that the lessonfile is ok.
    QSTATUS_NO mean that the question is corrupt and unusable. Other
               questions in the same file might be useable.
    """

    def __init__(self, exname):
        abstract.Teacher.__init__(self, exname)
        self.lessonfileclass = lessonfile.DictationLessonfile


class Gui(abstract.LessonbasedGui):

    def __init__(self, teacher):
        abstract.LessonbasedGui.__init__(self, teacher, no_notebook=False)
        ################
        # practise_box #
        ################
        self.g_question_title = gu.bLabel(self.practise_box, "", False, False)

        self.g_music_displayer = mpd.MusicDisplayer()
        self.set_size_request(500, -1)
        self.g_music_displayer.show()
        self.practise_box.pack_start(self.g_music_displayer, True, True, 0)
        ###############
        # action_area #
        ###############
        self.g_partbox = gu.bHBox(self.practise_box, False)
        self.g_go_back = Gtk.Button(stock='gtk-go-back')
        self.g_go_back.connect('clicked', self.select_previous)
        self.g_go_back.show()
        self.action_area.pack_start(self.g_go_back, False, False, 0)
        self.g_go_forward = Gtk.Button(stock='gtk-go-forward')
        self.g_go_forward.show()
        self.g_go_forward.connect('clicked', self.select_next)
        self.action_area.pack_start(self.g_go_forward, False, False, 0)
        self.g_play = gu.bButton(self.action_area, _("_Play the whole music"),
                                 self.play)
        self.g_show = gu.bButton(self.action_area, _("_Show"), self.show_answer)

    def exception_cleanup(self):
        """ cleanup function after exception caught in select_previous
        and select_next
        """
        soundcard.synth.stop()
        self.m_t.q_status = self.QSTATUS_NO
        self.g_play.set_sensitive(False)
        self.g_show.set_sensitive(False)
        self.g_music_displayer.clear()

    def select_previous(self, widget):
        self.m_t.m_P.select_previous()
        try:
            self.display_start_of_music()
        except Exception as e:
            if not self.standard_exception_handler(e, self.exception_cleanup):
                raise
        else:
            self.m_t.q_status = self.QSTATUS_NEW
            self.g_play.set_sensitive(True)
            self.g_show.set_sensitive(True)
        self._update()

    def select_next(self, widget):
        self.m_t.m_P.select_next()
        try:
            self.display_start_of_music()
        except Exception as e:
            if not self.standard_exception_handler(e, self.exception_cleanup):
                raise
        else:
            self.m_t.q_status = self.QSTATUS_NEW
            self.g_play.set_sensitive(True)
            self.g_show.set_sensitive(True)
        self._update()

    def play(self, widget=None):
        # see Teacher docstring.
        if self.m_t.q_status != self.QSTATUS_NEW:
            return
        try:
            self.m_t.m_P.play_question()
        except Exception as e:
            if not self.standard_exception_handler(e, soundcard.synth.stop):
                raise

    def on_end_practise(self):
        self.m_t.end_practise()

    def show_answer(self, widget=None):
        # see Teacher docstring.
        if self.m_t.q_status != self.QSTATUS_NEW:
            return
        try:
            self.g_music_displayer.display(self.m_t.m_P.get_music(),
                            self.get_int('config/feta_font_size=20'))
        except Exception as e:
            if not self.standard_exception_handler(e):
                raise

    def _update(self):
        """
        Updates the buttons above the action_area where you have
        one or more buttons with a small note pixmap on. Each of the
        buttons will play one part of the music in the question.
        """
        # tmp func used as callback function
        def f(w, start, end, self=self):
            try:
                utils.play_music(self.m_t.m_P.get_music(),
                    self.m_t.m_P.get_tempo(),
                    cfg.get_int('config/preferred_instrument'),
                    cfg.get_int('config/preferred_instrument_volume'),
                    start, end)
            except Exception as e:
                if not self.standard_exception_handler(e, soundcard.synth.stop):
                    raise
        for i in self.g_partbox.get_children():
            i.destroy()
        # if the lessonfile was invalid, m_P could be None
        if self.m_t.m_P and self.m_t.m_P.m_questions:
            if 'name' in self.m_t.m_P.get_question():
                self.g_question_title.set_text(self.m_t.m_P.get_name())
            else:
                self.g_question_title.set_text("")
            v = self.m_t.m_P.get_breakpoints()
            if v == []:
                # we display one button that will play the whole music if
                # there are not breakpoints in the music
                btn = self.create_pixmap_button()
                btn.connect('clicked', f, None, None)
                btn.show()
                self.g_partbox.pack_start(btn, True, True, 0)
                return
            tmp = [Rat(0, 1)] + v + [Rat(2**30, 1)]
            for i in range(len(tmp) - 1):
                btn = self.create_pixmap_button()
                btn.show()
                btn.connect('clicked', f, tmp[i], tmp[i + 1])
                self.g_partbox.pack_start(btn, True, True, 0)
        # q_status is QSTATUS_NO if the question is invalid (from the lessonfile)
        if self.m_t.q_status == self.QSTATUS_NO:
            self.g_partbox.set_sensitive(False)
        else:
            self.g_partbox.set_sensitive(True)

    def display_start_of_music(self):
        """
        Callers must catch exceptions.
        """
        fontsize = self.get_int('config/feta_font_size=20')
        try:
            if self.m_t.m_P.get_clue_music():
                self.g_music_displayer.display(self.m_t.m_P.get_clue_music().get_mpd_music_string(self.m_t.m_P), fontsize)
            elif self.m_t.m_P.get_clue_end():
                self.g_music_displayer.display(self.m_t.m_P.get_music(),
                            fontsize, self.m_t.m_P.get_clue_end())
            else:
                self.g_music_displayer.display(self.m_t.m_P.get_music(),
                            fontsize, Rat(0, 1))
        except mpd.MpdException as e:
            if self.m_t.m_P.get_clue_music():
                e.m_mpd_varname = 'clue_music'
            else:
                e.m_mpd_varname = 'music'
            self.m_t.m_P.get_question()['music'].complete_to_musicdata_coords(self.m_t.m_P, e)
            if 'm_mpd_badcode' not in dir(e):
                e.m_mpd_badcode = self.m_t.m_P.get_question()[e.m_mpd_varname].get_err_context(e, self.m_t.m_P)
            raise

    def update_gui_after_lessonfile_change(self):
        self.m_t.q_status = self.QSTATUS_NEW
        if not self.m_t.m_P.m_questions:
            solfege.win.display_error_message(_("The lesson file '%s' contains no questions.") % self.m_t.m_lessonfile)
            self.g_play.set_sensitive(False)
            self.g_show.set_sensitive(False)
            self.g_go_back.set_sensitive(False)
            self.g_go_forward.set_sensitive(False)
            self._update()
            return
        else:
            self.g_go_forward.set_sensitive(True)
            self.g_go_back.set_sensitive(True)
        self.m_t.m_P.select_first()
        self.action_area.set_sensitive(True)
        try:
            self.display_start_of_music()
        except Exception as e:
            def cleanup_function():
                self.m_t.q_status = self.QSTATUS_NO
                self.g_play.set_sensitive(False)
                self.g_show.set_sensitive(False)
                self.g_music_displayer.clear()
            if not self.standard_exception_handler(e, cleanup_function):
                raise
        else:
            self.g_play.set_sensitive(True)
            self.g_show.set_sensitive(True)
        self._update()

    def create_pixmap_button(self):
        im = Gtk.Image()
        im.set_from_stock("solfege-rhythm-c4", Gtk.IconSize.LARGE_TOOLBAR)
        im.show()
        btn = Gtk.Button()
        btn.add(im)
        return btn

    def on_start_practise(self):
        super(Gui, self).on_start_practise()
        self.update_gui_after_lessonfile_change()
        self.g_play.grab_focus()



# GNU Solfege - free ear training
# Copyright (C) 2009, 2011, 2016 Tom Cato Amundsen

from __future__ import absolute_import
from __future__ import print_function
import sys
print(sys.version)

from gi.repository import Gtk
print(Gtk)


class Win(Gtk.Window):

    def __init__(self):
        Gtk.Window.__init__(self)
        self.set_default_size(600, 400)
        b = Gtk.Button("Click to quit")
        b.connect('clicked', Gtk.main_quit)
        self.add(b)

w = Win()
w.show_all()
Gtk.main()



# vim: set fileencoding=utf-8 :
# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING


import unittest
import codecs
import doctest
import os.path
from solfege.testlib import outdir

import solfege.cfg
from solfege.cfg import parse_file_into_dict
from solfege import cfg


class Test_parse_file_into_dict(unittest.TestCase):

    def test_parse_default_config(self):
        d = {}
        parse_file_into_dict(d, "default.config")
        self.assertEqual(d['gui']['expert_mode'], "false")

    def test_fail_on_non_utf8(self):
        filename = os.path.join(outdir, 'bad-unicode.ini')
        outfile = open(filename, 'w', encoding="iso-8859-1")
        outfile.write("# This file is not in utf8 encoding\n"
             "[sound]\n"
             "midi_player=/ABC/Us\xe9r\xa0\xff\xff\xff\x00r /bin/timidity\n")
        outfile.close()
        d = {}
        self.assertRaises(UnicodeDecodeError, parse_file_into_dict, d, filename)
        os.remove(filename)

    def test_parse_utf8(self):
        filename = os.path.join(outdir, 'ok-utf8.ini')
        outfile = codecs.open(filename, 'w', 'utf-8')
        outfile.write("# This file is in utf8 encoding\n"
                      "[sound]\n"
                      "s=/home/Usér/bin/prog\n"
                      "f=1.1\n"
                      "i=3\n")
        outfile.close()
        d = {}
        parse_file_into_dict(d, filename)
        self.assertEqual(d['sound']['i'], '3')
        self.assertEqual(d['sound']['f'], '1.1')
        self.assertEqual(d['sound']['s'], '/home/Us\xe9r/bin/prog')
        self.assertTrue(isinstance(d['sound']['s'], str))
        self.assertTrue(isinstance(d['sound']['s'], str))

    def test_set_str(self):
        cfg.set_string("abc/def", "str string")
        cfg.set_string("abc/ghi", "unicode string ØÆÅ")
        self.assertTrue(isinstance(cfg.get_string("abc/def"), str))
        self.assertTrue(isinstance(cfg.get_string("abc/ghi"), str))

suite = unittest.makeSuite(Test_parse_file_into_dict)
suite.addTest(doctest.DocTestSuite(solfege.cfg))



# vim: set fileencoding=utf-8 :
# Solfege - free ear training software
# Copyright (C) 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING


import unittest

from solfege.dataparser import Dataparser
from solfege.lfmod import parse_tree_interpreter
from solfege.testlib import TmpFileBase
import solfege.parsetree as pt


class TestLfMod(TmpFileBase):
    parserclass = Dataparser

    def get_mod(self, builtins=None):
        return parse_tree_interpreter(self.p.tree, builtins)

    def test_simplest(self):
        self.do_file("a = 1  b = 5")
        mod = self.get_mod()
        self.assertEqual(mod.m_globals['a'], 1)
        self.assertEqual(mod.m_globals['b'], 5)

    def test_list_assignment(self):
        self.do_file("s = 3, 4, 5 ")
        self.assertEqual(self.get_mod().m_globals['s'], [3, 4, 5])

    def test_addition(self):
        self.do_file("a = 3 + 4")
        mod = self.get_mod()
        self.assertEqual(mod.m_globals['a'], 7)
        self.assertTrue(isinstance(self.p.tree[0], pt.Assignment))
        self.assertTrue(isinstance(self.p.tree[0].left, pt.Identifier))
        self.assertTrue(isinstance(self.p.tree[0].right, pt.Addition))

    def test_question(self):
        self.do_file("a = 2 question { q = 2 }")
        mod = self.get_mod()
        self.assertEqual(list(mod.m_globals.keys()), ['a'])
        self.assertEqual(mod.m_blocklists['question'][0]['q'], 2)

    def test_nested_block(self):
        """
        parse_tree_interpreter does not handle nested blocks, even though
        Dataparser and parse does.
        FIXME: an exception and error message that make sense for end
        users would be preferred.
        """
        self.do_file("question { a = 4 subbl { b = 5 } }")
        self.assertRaises(Exception, self.get_mod)

    def test_named_block(self):
        self.do_file("element I { q = 2 }")
        mod = self.get_mod()
        self.assertTrue('I' in mod.m_globals)
        self.assertEqual(mod.m_globals['I']['q'], 2)

    def test_global_lookup_in_question(self):
        self.do_file("\n".join([
            "a = 3",
            "question { q = a }",
        ]))
        mod = self.get_mod()
        self.assertEqual(mod.m_globals['a'], 3)
        self.assertEqual(mod.m_blocklists['question'][0]['q'], 3)

    def test_music_shortcut(self):
        def chord(s): return s
        self.do_file('question { name="test-name" chord("c e g") } ')
        mod = self.get_mod({'chord': (None, chord)})
        self.assertEqual(mod.m_blocklists['question'][0]['name'], "test-name")
        self.assertEqual(mod.m_blocklists['question'][0]['music'], "c e g")

    def test_global_lookup_in_question_redef_global(self):
        self.do_file("\n".join([
            "a = 3",
            "question { q = a }",
            "a = 4",
            "question { q = a }",
        ]))
        mod = self.get_mod()
        self.assertEqual(mod.m_blocklists['question'][0]['q'], 3)
        self.assertEqual(mod.m_blocklists['question'][1]['q'], 4)
        self.assertEqual(mod.m_globals['a'], 4)

    def test_1level_namespace_lookup(self):
        self.p.parse_file("solfege/tests/lesson-files/test_absolute_import_not_found_load_relative")
        mod = parse_tree_interpreter(self.p.tree)
        self.assertEqual(mod.m_globals['var'], 55)

suite = unittest.makeSuite(TestLfMod)



#!/usr/bin/python3

import os
import sys
# We need to do this on the build server. For some reason, it is not
# necessary on my workstation.
sys.path.insert(0, os.getcwd())

import solfege.const

img_str = """%i:<inlinemediaobject>
      <imageobject>
        <imagedata fileref="../../graphics/rhythm-%s.png" format="PNG"/>
      </imageobject>
      <textobject>
       <phrase>%s</phrase>
      </textobject>
    </inlinemediaobject>"""
f = open("help/C/rhythmtable.xml", "w")
print("<para>", file=f)
for i, r in enumerate(solfege.const.RHYTHMS):
    f.write(img_str % (i, r.replace(" ", ""), r))
    if i != len(solfege.const.RHYTHMS) - 1:
        print(", ", file=f)
print("</para>", file=f)
f.close()



#!/usr/bin/python3
# vim: set fileencoding=utf-8:
# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This script is used to launch Solfege when running
# it from the source dir without installing.

import builtins
import time
builtins.start_time = time.time()

import sys
import os
import os.path
import shutil

from solfege import cfg
from solfege import filesystem

if sys.platform == 'win32':
    # Migration added in solfege 3.9.0.
    try:
        if not os.path.exists(filesystem.app_data()):
            if os.path.exists(os.path.join(filesystem.get_home_dir(), ".solfege")):
                shutil.copytree(os.path.join(filesystem.get_home_dir(), ".solfege"),
                                filesystem.app_data())
            else:
                os.mkdir(filesystem.app_data())
        if not os.path.exists(filesystem.rcfile()):
            if os.path.exists(os.path.join(filesystem.get_home_dir(), ".solfegerc")):
                shutil.copy(os.path.join(filesystem.get_home_dir(), ".solfegerc"),
                            filesystem.rcfile())
    except (IOError, os.error) as e:
        print("Migration failed:", e)

from solfege import presetup
presetup.presetup("default.config", None, filesystem.rcfile())

# i18n should be imported very early in program init because it setup
# the _ and _i functions for the whole program.

from solfege import i18n
i18n.setup(".", cfg.get_string("app/lc_messages"))
import solfege.startup
solfege.startup.start_app(".")



#!/usr/bin/python

"""
This is a wrapper around the pydoc script. It is necessary to use this on
most of the files in solfege because they require i18n setup.
"""

import sys
import pydoc
sys.path.insert(0, ".")
import solfege.i18n
solfege.i18n.setup(".")

pydoc.cli()



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011,
# 2013  Tom Cato Amundsen
# Copyright (C) 2013 Jan Baumgart (Folkwang Universitaet der Kuenste)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk

from solfege import abstract
from solfege import abstract_solmisation_addon
from solfege import gu
from solfege import lessonfile
from solfege import mpd

from solfege.const import solmisation_syllables, solmisation_notenames


class Teacher(abstract.Teacher, abstract_solmisation_addon.SolmisationAddOnClass):
    OK = 0
    ERR_PICKY = 1
    ERR_NO_ELEMS = 2

    def __init__(self, exname):
        abstract.Teacher.__init__(self, exname)
        self.lessonfileclass = lessonfile.HeaderLessonfile

    def play_question(self):
        if self.q_status == self.QSTATUS_NO:
            return
        self.play(self.get_music_string())

    def guess_answer(self, a):
        assert self.q_status in [self.QSTATUS_NEW, self.QSTATUS_WRONG]
        v = []
        for idx in range(len(self.m_question)):
            v.append(self.m_question[idx] == a[idx])
        if not [x for x in v if x == 0]:
            self.q_status = self.QSTATUS_SOLVED
            self.maybe_auto_new_question()
            return 1
        else:
            self.q_status = self.QSTATUS_WRONG


class RhythmViewer(Gtk.Frame):

    def __init__(self, parent):
        Gtk.Frame.__init__(self)
        self.set_shadow_type(Gtk.ShadowType.IN)
        self.g_parent = parent
        self.g_box = Gtk.HBox()
        self.g_box.show()
        self.g_box.set_spacing(gu.PAD_SMALL)
        self.g_box.set_border_width(gu.PAD)
        self.add(self.g_box)
        self.m_data = []
        # the number of rhythm elements the viewer is supposed to show
        self.m_num_notes = 0
        self.g_face = None
        self.__timeout = None

    def set_num_notes(self, i):
        self.m_num_notes = i

    def clear(self):
        for child in self.g_box.get_children():
            child.destroy()
        self.m_data = []

    def create_holders(self):
        """
        create those |__| that represents one beat
        """
        if self.__timeout:
            GObject.source_remove(self.__timeout)
        self.clear()
        for x in range(self.m_num_notes):
            self.g_box.pack_start(gu.create_png_image('holder'), False, False, 0)
        self.m_data = []

    def clear_wrong_part(self):
        """When the user have answered the question, this method is used
        to clear all but the first correct elements."""
        # this assert is always true because if there is no rhythm element,
        # then there is a rhythm holder ( |__| )
        assert self.m_num_notes == len(self.g_parent.m_t.m_question)
        self.g_face.destroy()
        self.g_face = None
        for n in range(self.m_num_notes):
            if self.m_data[n] != self.g_parent.m_t.m_question[n]:
                break
        for x in range(n, len(self.g_box.get_children())):
            self.g_box.get_children()[n].destroy()
        self.m_data = self.m_data[:n]
        for x in range(n, self.m_num_notes):
            self.g_box.pack_start(gu.create_png_image('holder'), False, False, 0)

    def add_rhythm_element(self, i):
        assert len(self.m_data) <= self.m_num_notes
        if len(self.g_box.get_children()) >= self.m_num_notes:
            self.g_box.get_children()[self.m_num_notes - 1].destroy()
        vbox = Gtk.VBox()
        vbox.show()
#        im = gu.create_rhythm_image(const.RHYTHMS[i])
        im = self.g_parent.solbutton(i, False)
        vbox.pack_start(im, True, True, 0)
        vbox.pack_start(gu.create_png_image('rhythm-wrong'), False, False, 0)
        vbox.get_children()[-1].hide()
        self.g_box.pack_start(vbox, False, False, 0)
        self.g_box.reorder_child(vbox, len(self.m_data))
        self.m_data.append(i)

    def backspace(self):
        if len(self.m_data) > 0:
            if self.g_face:
                self.g_box.get_children()[-2].destroy()
                self.g_face.destroy()
                self.g_face = None
            self.g_box.get_children()[len(self.m_data) - 1].destroy()
            self.g_box.pack_start(gu.create_png_image('holder'), False, False, 0)
            del self.m_data[-1]

    def mark_wrong(self, idx):
        """
        Mark the rhythm elements that was wrong by putting the content of
        graphics/rhythm-wrong.png (normally a red line) under the element.
        """
        self.g_box.get_children()[idx].get_children()[1].show()

    def len(self):
        "return the number of rhythm elements currently viewed"
        return len(self.m_data)

    def sad_face(self):
        l = gu.HarmonicProgressionLabel(_("Wrong"))
        l.show()
        self.g_box.pack_start(l, False, False, 0)
        self.g_face = Gtk.EventBox()
        self.g_face.connect('button_press_event', self.on_sadface_event)
        self.g_face.show()
        im = Gtk.Image.new_from_icon_name(
            "face-sad", Gtk.IconSize.LARGE_TOOLBAR)
        im.show()
        self.g_face.add(im)
        self.g_box.pack_start(self.g_face, False, False, 0)

    def happy_face(self):
        l = gu.HarmonicProgressionLabel(_("Correct"))
        l.show()
        self.g_box.pack_start(l, False, False, 0)
        self.g_face = Gtk.EventBox()
        self.g_face.connect('button_press_event', self.on_happyface_event)
        self.g_face.show()
        im = Gtk.Image.new_from_icon_name(
            "face-smile", Gtk.IconSize.LARGE_TOOLBAR)
        im.show()
        self.g_face.add(im)
        self.g_box.pack_start(self.g_face, False, False, 0)

    def on_happyface_event(self, obj, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 1:
            self.g_parent.new_question()

    def on_sadface_event(self, obj, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 1:
            self.clear_wrong_part()

    def flash(self, s):
        self.clear()
        l = Gtk.Label(label=s)
        l.set_name("Feedback")
        l.set_alignment(0.0, 0.5)
        l.show()
        self.g_box.pack_start(l, True, True, 0)
        self.g_box.set_size_request(
            max(l.size_request().width + gu.PAD * 2, self.g_box.size_request().width),
            max(l.size_request().height + gu.PAD * 2, self.g_box.size_request().height))
        self.__timeout = GObject.timeout_add(2000, self.unflash)

    def unflash(self, *v):
        self.__timeout = None
        self.clear()


class Gui(abstract.Gui, abstract_solmisation_addon.SolmisationAddOnGuiClass):
    lesson_heading = _("Solmisation Diktat")

    def __init__(self, teacher):
        abstract.Gui.__init__(self, teacher)
        self.m_key_bindings = {'backspace_ak': self.on_backspace}
        self.g_answer_box = Gtk.VBox()
        self.answer_buttons = []
        self.m_answer_buttons = {}

        #-------
        hbox = gu.bHBox(self.practise_box)
        b = Gtk.Button(_("Play"))
        b.show()
        b.connect('clicked', self.play_users_answer)
        hbox.pack_start(b, False, True, 0)
        self.practise_box.pack_start(Gtk.HBox(), False, False,
                                     padding=gu.PAD_SMALL)
        self.g_rhythm_viewer = RhythmViewer(self)
        self.g_rhythm_viewer.create_holders()
        hbox.pack_start(self.g_rhythm_viewer, True, True, 0)

        self.practise_box.pack_start(self.g_answer_box, False, False, 0)

        # action area
        self.std_buttons_add(
            ('new', self.new_question),
            ('repeat', self.repeat_question),
            #('play_answer', self.play_users_answer),
            ('give_up', self.give_up),
            ('backspace', self.on_backspace))

        self.practise_box.show_all()
        ###############
        # config_grid #
        ###############
        self.add_select_elements_gui(row=0)
        #--------
        self.add_select_num_notes_gui(row=1)
        #-----
        self.g_config_grid.attach(Gtk.Label(_("Beats per minute:")),
            0, 2, 1, 1)
        spin = gu.nSpinButton(self.m_exname, 'bpm',
                 Gtk.Adjustment(60, 20, 240, 1, 10))
        self.g_config_grid.attach(spin, 1, 2, 1, 1)
        self.g_config_grid.attach(
           gu.nCheckButton(self.m_exname,
                           "show_first_note",
                           _("Show the first tone")),
           0, 3, 1, 1)
        self.g_config_grid.attach(gu.nCheckButton(self.m_exname,
                                        "play_cadence",
                                        _("Play cadence")),
                                  0, 4, 1, 1)
        self._add_auto_new_question_gui(row=5)
        self.g_config_grid.show_all()

    def solbutton(self, i, connect):
        if i > len(solmisation_syllables) or i < 0:
            btn = Gtk.Button()
        else:
            btn = Gtk.Button(solmisation_syllables[i])
        btn.show()
        if connect:
            btn.connect('clicked', self.guess_element, i)
        return btn

    def select_element_cb(self, button, element_num):
        super(Gui, self).select_element_cb(button, element_num)
        self.m_answer_buttons[element_num].set_sensitive(button.get_active())
        # self.update_answer_buttons()

    def on_backspace(self, widget=None):
        if self.m_t.q_status == self.QSTATUS_SOLVED:
            return
        self.g_rhythm_viewer.backspace()
        if not self.g_rhythm_viewer.m_data:
            self.g_backspace.set_sensitive(False)

    def play_users_answer(self, widget):
        if self.g_rhythm_viewer.m_data:
            melody = ""
            for k in self.g_rhythm_viewer.m_data:
                melody += " " + mpd.transpose_notename(solmisation_notenames[k], self.m_t.m_transp) + "4"

            self.m_t.play(r"\staff{ \time 1000000/4 %s }" % melody)

    def guess_element(self, sender, i):
        if self.m_t.q_status == self.QSTATUS_NO:
            self.g_rhythm_viewer.flash(_("Click 'New' to begin."))
            return
        if self.m_t.q_status == self.QSTATUS_SOLVED:
            return
        if self.g_rhythm_viewer.len() == len(self.m_t.m_question):
            self.g_rhythm_viewer.clear_wrong_part()
        self.g_backspace.set_sensitive(True)
        self.g_rhythm_viewer.add_rhythm_element(i)
        if self.g_rhythm_viewer.len() == len(self.m_t.m_question):
            if self.m_t.guess_answer(self.g_rhythm_viewer.m_data):
                self.g_rhythm_viewer.happy_face()
                self.std_buttons_answer_correct()
            else:
                v = []
                for idx in range(len(self.m_t.m_question)):
                    v.append(self.m_t.m_question[idx] == self.g_rhythm_viewer.m_data[idx])
                for x in range(len(v)):
                    if not v[x]:
                        self.g_rhythm_viewer.mark_wrong(x)
                self.g_rhythm_viewer.sad_face()
                self.std_buttons_answer_wrong()

    def new_question(self, widget=None):
        g = self.m_t.new_question()
        if g == self.m_t.OK:
            self.g_first_rhythm_button.grab_focus()
            self.g_rhythm_viewer.set_num_notes(self.get_int('num_notes'))
            self.g_rhythm_viewer.create_holders()
            self.std_buttons_new_question()
            if self.m_t.get_bool('show_first_note'):
                self.g_rhythm_viewer.add_rhythm_element(self.m_t.m_question[0])
            try:
                self.m_t.play_question()
            except Exception as e:
                if not self.standard_exception_handler(e, self.on_end_practise):
                    raise
                return
        elif g == self.m_t.ERR_PICKY:
            self.g_rhythm_viewer.flash(_("You have to solve this question first."))
        else:
            assert g == self.m_t.ERR_NO_ELEMS
            self.g_repeat.set_sensitive(False)
            self.g_rhythm_viewer.flash(_("You have to configure this exercise properly"))

    def repeat_question(self, *w):
        self.m_t.play_question()
        self.g_first_rhythm_button.grab_focus()

    def update_answer_buttons(self):
        """
        (Re)create the buttons needed to answer the questions.
        We recreate the buttons for each lesson file because the
        header may specify a different set of rhythm elements to use.
        """
        for but in self.answer_buttons:
            but.destroy()
        self.answer_buttons = []
        self.g_first_rhythm_button = None

        gs = Gtk.SizeGroup(Gtk.SizeGroupMode.HORIZONTAL)

        for i, v in enumerate((
                    [1, 4, -1, 8, 11, -1, 15, 18, 21, -1, 25, 28, -1, 32],
                    [0, 3, 6, 7, 10, 13, 14, 17, 20, 23, 24, 27, 30, 31, 34],
                    [2, 5, -1, 9, 12, -1, 16, 19, 22, -1, 26, 29, -1, 33])):
            hbox = Gtk.HBox(True, 0)
            for k in v:
                b = self.solbutton(k, True)
                gs.add_widget(b)
                b.set_sensitive(False)
                for n in self.m_t.m_P.header.solmisation_elements:
                    if k == n:
                        b.set_sensitive(True)
                        if not self.g_first_rhythm_button:
                            self.g_first_rhythm_button = b
                hbox.pack_start(b, True, True, 0)
                self.answer_buttons.append(b)
                if k != -1:
                    self.m_answer_buttons[k] = b
            spacing = Gtk.Alignment()
            if i in (0, 2):
                spacing.set_property('left-padding', 16)
                spacing.set_property('right-padding', 16)
            spacing.add(hbox)
            self.g_answer_box.pack_start(spacing, True, True, 0)
            spacing.show_all()

    def on_start_practise(self):
        # FIXME for now, we run in custom_mode all the time, so we don't
        # have to add lots of lesson files. We can change this later.
        #self.m_t.m_custom_mode = self.get_bool('gui/expert_mode')
        self.m_t.m_custom_mode = True
        super(Gui, self).on_start_practise()
        if not self.m_t.m_P.header.solmisation_elements:
            self.m_t.m_P.header.solmisation_elements = self.m_t.elements[:]
        self.update_answer_buttons()
        self.std_buttons_start_practise()
        if self.m_t.m_custom_mode:
            self.update_select_elements_buttons()
            self.g_element_frame.show()
        else:
            self.g_element_frame.hide()
        self.m_t.set_default_header_values()
        if 'show_first_note' in self.m_t.m_P.header:
            self.m_t.set_bool('show_first_note', self.m_t.m_P.header.show_first_note)
        if 'play_cadence' in self.m_t.m_P.header:
            self.m_t.set_bool('play_cadence', self.m_t.m_P.header.play_cadence)
        self.g_rhythm_viewer.flash(_("Click 'New' to begin."))

    def on_end_practise(self):
        self.m_t.end_practise()
        self.std_buttons_end_practise()
        self.g_rhythm_viewer.create_holders()

    def give_up(self, widget=None):
        if self.m_t.q_status == self.QSTATUS_NO:
            return
        self.g_rhythm_viewer.clear()
        for i in self.m_t.m_question:
            self.g_rhythm_viewer.add_rhythm_element(i)
        self.m_t.q_status = self.QSTATUS_SOLVED
        self.std_buttons_give_up()



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys

from gi.repository import Gtk

import solfege
from solfege import gu


class TracebackWindow(Gtk.Dialog):

    RESPONSE_CLEAR = 1
    def __init__(self, show_gtk_warnings):
        Gtk.Dialog.__init__(self)
        self.m_show_gtk_warnings = show_gtk_warnings
        self.set_default_size(630, 400)
        self.vbox.set_border_width(8)
        label = Gtk.Label(label=_("GNU Solfege message window"))
        label.set_name('Heading2')
        self.vbox.pack_start(label, False, False, 0)
        label = Gtk.Label(label=_("Please report this to the bug database or send an email to bug-solfege@gnu.org if the content of the message make you believe you have found a bug."))
        label.set_line_wrap(True)
        self.vbox.pack_start(label, False, False, 0)
        scrollwin = Gtk.ScrolledWindow()
        scrollwin.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.vbox.pack_start(scrollwin, True, True, 0)
        self.g_text = Gtk.TextView()
        scrollwin.add(self.g_text)
        self.add_button("gtk-clear", self.RESPONSE_CLEAR)
        self.add_button("gtk-close", Gtk.ResponseType.CLOSE)
        self.connect('response', self.on_response)

    def write(self, txt):
        if getattr(solfege, 'win', False):
            self.set_transient_for(solfege.win)
        if ("DeprecationWarning:" in txt) or \
           (not self.m_show_gtk_warnings and (
            "GtkWarning" in txt
            or "PangoWarning" in txt
            or ("Python C API version mismatch" in txt and
                ("solfege_c_midi" in txt or "swig" in txt))
            )):
            return
        sys.stdout.write(txt)
        if txt.strip():
            self.show_all()
        buffer = self.g_text.get_buffer()
        buffer.insert(buffer.get_end_iter(), txt)
        self.set_focus(
            self.get_widget_for_response(Gtk.ResponseType.CLOSE))

    def flush(self, *v):
        pass

    def close(self, *v):
        pass

    def on_response(self, dlg, response):
        if response == self.RESPONSE_CLEAR:
            b = self.g_text.get_buffer()
            b.delete(b.get_start_iter(), b.get_end_iter())
        elif response == Gtk.ResponseType.CLOSE:
            self.hide()



# vim: set fileencoding=utf-8 :
# GNU Solfege - free ear training software
# Copyright (C) 2010, 2011, 2016 Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import shutil

from gi.repository import GObject
from gi.repository import Gtk

from solfege import filesystem
from solfege import gu


class NewProfileDialog(Gtk.Dialog):

    def __init__(self):
        Gtk.Dialog.__init__(self, _("_Create profile\u2026").replace("\u2026", "").replace("_", ""))
        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                         Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT)
        vbox = gu.hig_dlg_vbox()
        self.vbox.pack_start(vbox, True, True, 0)
        #
        label = Gtk.Label(label=_("Enter the name of the new folder:"))
        label.set_alignment(0.0, 0.5)
        vbox.pack_start(label, False, False, 0)
        #
        self.g_entry = Gtk.Entry()
        self.g_entry.connect('changed', self.on_entry_changed)
        self.g_entry.set_activates_default(True)
        vbox.pack_start(self.g_entry, False, False, 0)
        #
        label = Gtk.Label(label=_("Your profile data will be stored in:"))
        label.set_alignment(0.0, 0.5)
        vbox.pack_start(label, False, False, 0)
        #
        self.g_profile_location = Gtk.Label()
        vbox.pack_start(self.g_profile_location, False, False, 0)
        #
        self.g_statusbox = Gtk.HBox()
        self.g_statusbox.set_no_show_all(True)
        vbox.pack_start(self.g_statusbox, False, False, 0)
        im = Gtk.Image()
        im.set_from_stock(Gtk.STOCK_DIALOG_WARNING, Gtk.IconSize.MENU)
        self.g_statusbox.pack_start(im, False, False, 0)
        im.show()

        self.g_status = Gtk.Label()
        self.g_status.show()
        self.g_statusbox.pack_start(self.g_status, False, False, 0)
        self.g_entry.set_text(_("New Profile"))
        self.set_default_response(Gtk.ResponseType.ACCEPT)

    def on_entry_changed(self, *w):
        pdir = os.path.join(filesystem.app_data(), "profiles",
                            self.g_entry.get_text())
        self.g_profile_location.set_text(pdir)
        if os.path.exists(pdir):
            self.g_status.set_text(_("The profile «%s» already exists.") % self.g_entry.get_text())
            self.g_statusbox.show()
            self.set_response_sensitive(Gtk.ResponseType.ACCEPT, False)
        else:
            self.g_statusbox.hide()
            self.g_status.set_text("")
            self.set_response_sensitive(Gtk.ResponseType.ACCEPT, True)


class RenameProfileDialog(Gtk.Dialog):

    def __init__(self, oldname):
        Gtk.Dialog.__init__(self, _("_Rename profile\u2026").replace("_", "").replace("\u2026", ""))
        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                        Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT)
        vbox = gu.hig_dlg_vbox()
        self.vbox.pack_start(vbox, True, True, 0)

        label = Gtk.Label(label=_("Rename the profile «%s» to:") % oldname)
        label.set_alignment(0.0, 0.5)
        vbox.pack_start(label, False, False, 0)

        self.g_entry = Gtk.Entry()
        self.g_entry.set_text(oldname)
        self.g_entry.set_activates_default(True)
        self.g_entry.connect('changed', self.on_entry_changed)
        vbox.pack_start(self.g_entry, False, False, 0)

        self.g_info = Gtk.Label()
        self.g_info.set_no_show_all(True)
        vbox.pack_start(self.g_info, False, False, 0)
        self.set_default_response(Gtk.ResponseType.ACCEPT)

    def on_entry_changed(self, w):
        s = self.g_entry.get_text()
        pdir = os.path.join(filesystem.app_data(), "profiles", s)
        ok = False
        if not s:
            self.g_info.show()
            self.g_info.set_text("Empty string not allowed")
        elif os.path.exists(pdir):
            self.g_info.show()
            self.g_info.set_text(_("The profile «%s» already exists.") % self.g_entry.get_text())
        else:
            self.g_info.hide()
            ok = True
        self.set_response_sensitive(Gtk.ResponseType.ACCEPT, ok)


class ProfileManagerBase(Gtk.Dialog):

    def __init__(self, parent, default_profile):
        Gtk.Dialog.__init__(self, _("GNU Solfege - Choose User Profile"),
                            parent)
        # Set a small size, and let the widgets expand the dialog
        self.set_default_size(100, 100)
        # We save the initially selected profile, because we need to keep
        # track of it if the user renames it and then presses cancel.
        self.m_default_profile = default_profile
        vbox = gu.hig_dlg_vbox()
        self.get_content_area().pack_start(vbox, False, False, 0)
        l = Gtk.Label(_("Solfege will save your statistics and test results in the user profile. By adding additional user profiles to Solfege, multiple users can share a user account on the operating system."))
        l.set_alignment(0.0, 0.5)
        l.set_line_wrap(True)
        vbox.pack_start(l, True, True, 0)

        hbox = Gtk.HBox()
        hbox.set_spacing(gu.hig.SPACE_MEDIUM)
        vbox.pack_start(hbox, True, True, 0)
        button_box = Gtk.VBox()

        self.g_create_profile = Gtk.Button.new_with_mnemonic(_("_Create profile\u2026"))
        self.g_create_profile.connect('clicked', self.on_create_profile)
        button_box.pack_start(self.g_create_profile, False, False, 0)

        self.g_rename_profile = Gtk.Button.new_with_mnemonic(_("_Rename profile\u2026"))
        self.g_rename_profile.connect('clicked', self.on_rename_profile)
        button_box.pack_start(self.g_rename_profile, False, False, 0)

        self.g_delete_profile = Gtk.Button.new_with_mnemonic(_("_Delete profile\u2026"))
        self.g_delete_profile.connect('clicked', self.on_delete_profile)
        button_box.pack_start(self.g_delete_profile, False, False, 0)

        hbox.pack_start(button_box, False, False, 0)
        self.g_liststore = liststore = Gtk.ListStore(GObject.TYPE_STRING)
        liststore.append((_("Standard profile"),))
        if os.path.exists(os.path.join(filesystem.app_data(), 'profiles')):
            for subdir in os.listdir(os.path.join(filesystem.app_data(),
                'profiles')):
                liststore.append((subdir,))
        #
        self.g_tw = tw = Gtk.TreeView(liststore)
        tw.connect('row-activated', lambda a, b, c: self.response(Gtk.ResponseType.ACCEPT))
        tw.set_headers_visible(False)
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(None, renderer, text=0)
        tw.append_column(column)
        hbox.pack_start(tw, False, False, 0)
        tw.show()
        tw.connect('cursor-changed', self.on_cursor_changed)
        tw.set_cursor((0,))
        for idx, s in enumerate(self.g_liststore):
            if s[0] == default_profile:
                tw.set_cursor((idx, ))
        #
        chk = gu.nCheckButton("app", "noprofilemanager", _("D_on't ask at startup"))
        vbox.pack_start(chk, False, False, 0)
        self.show_all()

    def on_create_profile(self, w):
        dlg = NewProfileDialog()
        dlg.show_all()
        ret = dlg.run()
        if ret == Gtk.ResponseType.ACCEPT:
            pdir = os.path.join(filesystem.app_data(),
                                "profiles", dlg.g_entry.get_text())
            if not os.path.exists(pdir):
                try:
                    os.makedirs(pdir)
                    self.g_liststore.append((dlg.g_entry.get_text(),))
                    self.g_tw.set_cursor((len(self.g_liststore) - 1,))
                except OSError as e:
                    gu.display_exception_message(e)
        dlg.destroy()

    def on_rename_profile(self, w):
        if self.m_default_profile == self.get_profile():
            rename_default = True
        else:
            rename_default = False
        dlg = RenameProfileDialog(self.get_profile())
        dlg.show_all()
        ret = dlg.run()
        if ret == Gtk.ResponseType.ACCEPT:
            path, column = self.g_tw.get_cursor()
            try:
                os.rename(os.path.join(
                    filesystem.app_data(), "profiles", self.get_profile()),
                    os.path.join(filesystem.app_data(),
                        "profiles", dlg.g_entry.get_text()))
                if rename_default:
                    self.m_default_profile = dlg.g_entry.get_text()
            except OSError as e:
                gu.display_exception_message(e)
                dlg.destroy()
                return
            path, column = self.g_tw.get_cursor()
            self.g_liststore.set(self.g_liststore.get_iter(path),
                0, dlg.g_entry.get_text())
        dlg.destroy()

    def on_delete_profile(self, w):
        if gu.dialog_yesno(_("Permanently delete the user profile «%s»?") % self.get_profile(), self):
            path, column = self.g_tw.get_cursor()
            it = self.g_liststore.get_iter(path)
            try:
                shutil.rmtree(os.path.join(filesystem.app_data(), "profiles", self.get_profile()))
            except OSError as e:
                gu.display_exception_message(e)
                return
            self.g_liststore.remove(it)
            if not self.g_liststore.iter_is_valid(it):
                it = self.g_liststore[-1].iter
            self.g_tw.set_cursor(self.g_liststore.get_path(it))

    def on_cursor_changed(self, treeview):
        path, column = self.g_tw.get_cursor()
        if path:
            self.g_delete_profile.set_sensitive(list(path) != [0])
            self.g_rename_profile.set_sensitive(list(path) != [0])

    def get_profile(self):
        """
        Return None if the standard profile is selected.
        Return the directory name for other profiles.
        """
        cursor = self.g_tw.get_cursor()
        if list(cursor[0]) == [0]:
            return None
        it = self.g_liststore.get_iter(cursor[0])
        return self.g_liststore.get(it, 0)[0]


class ProfileManager(ProfileManagerBase):

    def __init__(self, parent, default_profile):
        ProfileManagerBase.__init__(self, parent, default_profile)
        self.add_button(Gtk.STOCK_QUIT, Gtk.ResponseType.CLOSE)
        b = self.add_button(_("_Start GNU Solfege"), Gtk.ResponseType.ACCEPT)
        b.grab_focus()
        self.set_default_response(Gtk.ResponseType.ACCEPT)


class ChangeProfileDialog(ProfileManagerBase):

    def __init__(self, parent, default_profile):
        ProfileManagerBase.__init__(self, parent, default_profile)
        self.add_button(Gtk.STOCK_APPLY, Gtk.ResponseType.ACCEPT)



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
The Interval class is a class that is used to do math with
intervals and musical pitches. You use this class if you
need to know the difference between an augmented second and
a minor third. If you don't require this you can probably
ise the simpler function mpdutils.int_to_intervalname
"""

import re
from solfege.mpd import _exceptions

# We have to list all possible interval names here so that xgettext
# can find them and add them to the .pot file for translation. We
# cannot translate the interval quality and interval number separately
# because we don't know how to concat them on other languages.
if 0:
    _("Perfect Unison")
    _("Diminished Unison")
    _("Augmented Unison")
    #
    _("Diminished Second")
    _("Minor Second")
    _("Major Second")
    _("Augmented Second")
    #
    _("Diminished Third")
    _("Minor Third")
    _("Major Third")
    _("Augmented Third")
    #
    _("Diminished Fourth")
    _("Perfect Fourth")
    _("Augmented Fourth")
    #
    _("Diminished Fifth")
    _("Perfect Fifth")
    _("Augmented Fifth")
    #
    _("Diminished Sixth")
    _("Minor Sixth")
    _("Major Sixth")
    _("Augmented Sixth")
    #
    _("Diminished Seventh")
    _("Minor Seventh")
    _("Major Seventh")
    _("Augmented Seventh")
    #
    _("Diminished Octave")
    _("Perfect Octave")
    _("Augmented Octave")
    # translators: Only translate the word after interval| , and don't include
    # interval| in the translated string. So for Norwegians, translate
    # "interval|Diminished" to "Forminsket". Do similar for all strings
    # that are preceded with "interval|"
    _("Minor Ninth")
    _("Major Ninth")
    _("Minor Tenth")
    _("Major Tenth")
    _("interval|Unison")
    _("interval|Second")
    _("interval|Third")
    _("interval|Fourth")
    _("interval|Fifth")
    _("interval|Sixth")
    _("interval|Seventh")
    _("interval|Octave")
    _("interval|Ninth")
    _("interval|Tenth")
    #
    _("interval|Diminished")
    _("interval|Perfect")
    _("interval|Augmented")
    _("interval|Minor")
    _("interval|Major")
    _("interval|Doubly-diminished")
    _("interval|Doubly-augmented")


min_interval = -24
max_interval = 24

short_name = (
    # translators: Sorry for changing a few of these interval names.
    # I want to follow the naming used by wikipedia an others.
    # See http://www.solfege.org/translate for important details
    # These strings are used as button labels for the check buttons
    # selecting which intervals to ask for.
    _i("interval|P1"),
    # translators: m2 is a short version of minor second
    _i("interval|m2"),
    # translators: M2 is a short string for major second
    _i("interval|M2"),
    # translators: m3 is a short string for minor third
    _i("interval|m3"),
    # translators: M3 is a short string for major third
    _i("interval|M3"),
    # translators: 4 is a short string for perfect fourth
    _i("interval|4"),
    # translators: TT is a short string for tritone
    _i("interval|TT"),
    # translators: 5 is a short string for perfect fifth
    _i("interval|P5"),
    # translators: m6 is a short string for minor sixth
    _i("interval|m6"),
    # translators: M6 is a short string for major sixth
    _i("interval|M6"),
    # translators: m7 is a short string for minor seventh
    _i("interval|m7"),
    # translators: M7 is a short string for major seventh
    _i("interval|M7"),
    # translators: 8 is a short string for perfect octave
    _i("interval|P8"),
    # translators: m9 is a short string for minor ninth
    _i("interval|m9"),
    # translators: M9 is a short string for major ninth
    _i("interval|M9"),
    # translators: m10 is a short string for minor tenth
    _i("interval|m10"),
    # translators: M10 is a short string for major tenth
    _i("interval|M10"),
    # translators: P11 is a short string for perfect 11th (octave + fourth)
    _i("interval|P11"),
    # translators: d12 is a short string for octave + tritone
    # I considered the name TT8 (tritone+octave), but rejected it.
    _i("interval|d12"),
    # translators: P12 is a short string for perfect 12th (octave + fifth)
    _i("interval|P12"),
    # translators: m13 is a short string for minor 13th (octave + m6)
    _i("interval|m13"),
    # translators: M13 is a short string for major 13th (octave + M6)
    _i("interval|M13"),
    # translators: m14 is a short string for minor 14th (octave + m7)
    _i("interval|m14"),
    # translators: M14 is a short string for major 14th (octave + M7)
    _i("interval|M14"),
    # translators: P15 is a short string for perfect double octave
    _i("interval|P15"),
    )


def number_name(steps):
    try:
        return {
            1: "Unison",
            2: "Second",
            3: "Third",
            4: "Fourth",
            5: "Fifth",
            6: "Sixth",
            7: "Seventh",
            8: "Octave",
            9: "Ninth",
            10: "Tenth",
            11: "Eleventh",
            12: "Twelfth",
            13: "Thirteenth",
            14: "Fourteenth",
            15: "Double Octave"}[steps]
    except KeyError:
        return "%ith" % steps


class InvalidIntervalnameException(_exceptions.MpdException):

    def __init__(self, notename):
        _exceptions.MpdException.__init__(self)
        self.m_intervalname = notename

    def __str__(self):
        return _("Invalid interval name: %s") % self.m_intervalname


class Interval:
    """
    The interval is internally a interval less than octave
    pluss n octaves. The data variables:
      m_dir
      m_octave
      m_interval
      m_mod
    should NOT be touched by anyone, except MusicalPitch.__add__
    """
    _nn_to_interval_quality = {
        'p': 'Perfect',
        'dd': 'Doubly-diminished',
        'd': 'Diminished',
        'm': 'Minor',
        'M': 'Major',
        'a': 'Augmented',
        'aa': 'Doubly-augmented',
        }

    def __init__(self, iname=None):
        self.m_dir = 1  # value as to be 1 or -1 for initialised obj
        self.m_octave = 0  # 0, 1, 2, 3 etc
        self.m_interval = 0  # 0:unison, 1:seond, ... 6: septim
        # unison:              dim   perfect   aug
        # second:  -2:dim -1:minor 0:major   1:aug
        # third:      dim    minor   major     aug
        # fourth:              dim   perfect   aug
        # fifth:               dim   perfect   aug
        # sixth:      dim    minor   major     aug
        # seventh:    dim    minor   major     aug
        self.m_mod = 0
        if iname:
            self.set_from_string(iname)

    def nn_to_translated_quality(interval_quality):
        """
        Return translated interval quality from internal short string.
        interval_quality can be: dd, d, m, M, a, a, p
        The C locale will return english names, as 'Perfect' and 'Diminished'
        """
        # Hack, just to xgettext should not grab the string for translation
        xgettext_wont_find_us = _i
        return xgettext_wont_find_us("interval|%s" % Interval._nn_to_interval_quality[interval_quality].title())
    nn_to_translated_quality = staticmethod(nn_to_translated_quality)

    def errorcheck(self):
        assert 0 <= self.m_interval <= 6
        assert -2 <= self.m_mod <= 1  # should be increased to -3 <= x <= 2
        assert self.m_octave >= 0
        assert self.m_dir in (-1, 1)

    def _set(self, direction, interval, mod, octave):
        self.m_dir = direction
        self.m_interval = interval
        self.m_mod = mod
        self.m_octave = octave
        if __debug__:
            self.errorcheck()
        return self

    def new_from_int(i):
        assert isinstance(i, int)
        new_int = Interval()
        new_int.set_from_int(i)
        return new_int
    new_from_int = staticmethod(new_from_int)

    def set_from_int(self, i):
        """It returns self to allow chaining: set_from_int(4).pretty_name()
        """
        if i < 0:
            self.m_dir = -1
        else:
            self.m_dir = 1
        self.m_octave = abs(i) // 12
        self.m_mod, self.m_interval = (
               (0, 0),          # unison
               (-1, 1), (0, 1),  # second
               (-1, 2), (0, 2),  # third
               (0, 3),          # fourth
               (-1, 4),         # dim 5, tritone
               (0, 4),          # fifth
               (-1, 5), (0, 5),  # sixth
               (-1, 6), (0, 6))[abs(i) % 12]  # seventh
        return self

    def set_from_string(self, s):
        """
        unison  p1
        second  m2 M2
        third   m3 M3
        fourth  p4
        fifth   d5 5
        sixth   m6 M6
        seventh m7 M7
        octave  p8
        ninth    m9 M9
        tenth   m10 M10
        """
        # up or down
        s_orig = s[:]
        s = s.strip()
        if s[0] == "-":
            self.m_dir = -1
            s = s[1:]
        else:
            self.m_dir = 1
        m = re.match("(m|M|d|a|p)(\d+)", s)
        if not m:
            raise InvalidIntervalnameException(s_orig)
        modifier, i = m.groups()
        i = int(i)
        if i <= 7:
            self.m_octave = 0
        else:
            self.m_octave = (i - 1) // 7
        self.m_interval = i - 1 - self.m_octave * 7
        if self.m_interval in (1, 2, 5, 6):
            try:
                self.m_mod = {'d': -2, 'm': -1, 'M': 0, 'a': 1}[modifier]
            except:
                raise InvalidIntervalnameException(s_orig)
        elif self.m_interval in (0, 3, 4):
            try:
                self.m_mod = {'d': -1, 'p': 0, '': 0, 'a': 1}[modifier]
            except:
                raise InvalidIntervalnameException(s_orig)

    def get_intvalue(self):
        if __debug__:
            self.errorcheck()
        return ([0, 2, 4, 5, 7, 9, 11][self.m_interval] + self.m_octave * 12 + self.m_mod) * self.m_dir

    def __str__(self):
        if __debug__:
            self.errorcheck()
        ret = "(Interval %i %imod %io" % (self.m_interval, self.m_mod, self.m_octave)
        if self.m_dir == -1:
            ret = ret + " down)"
        else:
            ret = ret + " up)"
        return ret

    def __repr__(self):
        if self.m_interval in (0, 3, 4):
            return "%s%s" % ({-2: 'dd', -1: 'd', 0: 'p', 1: 'a', 2: 'aa'}[self.m_mod], self.m_interval + 1 + self.m_octave * 7)
        return "%s%s" % ({-2: 'd', -1: 'm', 0: 'M', 1: 'a'}[self.m_mod], (self.m_interval + 1 + self.m_octave * 7))

    def __eq__(self, interval):
        return self.m_dir == interval.m_dir \
            and self.m_octave == interval.m_octave \
            and self.m_mod == interval.m_mod \
            and self.m_interval == interval.m_interval

    def get_number_name(self):
        """
        Return the translated general name of the interval, like second, third
        etc. (major, minor etc.)
        """
        return _(number_name(self.steps()))

    def get_quality_short(self):
        """
        Return a short string telling the quality.
        This is a non-translated short string mostly used
        internally in the program.
        """
        if self.m_interval in (0, 3, 4):
            return {-2: "dd",
                    -1: "d",
                     0: "p",
                     1: "a",
                     2: "aa"}[self.m_mod]
        else:
            assert self.m_interval in (1, 2, 5, 6)
            return {-2: "d",
                    -1: "m",
                     0: "M",
                     1: "a"}[self.m_mod]

    def get_cname_short(self):
        """
        Return the short untranslated interval name, like p5 or M2
        """
        return "%s%i" % (self.get_quality_short(), self.steps())

    def get_cname(self):
        """
        Return the full untranslated interval name, both the number and quality.
        """
        if self.m_interval == 4 and self.m_mod == -1:
            if self.m_octave == 0:
                return "Tritone"
            elif self.m_octave == 1:
                return "Octave + Tritone"
        return "%s %s" % (
                self._nn_to_interval_quality[self.get_quality_short()],
                number_name(self.steps()))

    def get_name(self):
        """
        Return the full translated intervalname, both the number and quality.
        """
        return _(self.get_cname())

    def steps(self):
        return self.m_octave * 7 + self.m_interval + 1



#!/usr/bin/python
# GNU Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os.path
import sys
import textwrap
import traceback

import gi
# Gtk win windows complains without this:
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from solfege import cfg
from solfege import filesystem
from solfege import i18n


def presetup(app_defaults_filename, system_filename, user_filename):
    if not os.path.exists(filesystem.app_data()):
        os.makedirs(filesystem.app_data())
    if not os.path.exists(filesystem.user_data()):
        os.makedirs(filesystem.user_data())
    try:
        cfg.initialise(app_defaults_filename, system_filename, user_filename)
    except UnicodeDecodeError:
        traceback.print_exc()
        print(file=sys.stderr)
        print("\n".join(textwrap.wrap(
              "Your %s file is not properly utf8 encoded. Most likely"
              " it is the path to some external program that contain non-ascii"
              " characters. Please edit or delete the file. Or email it to"
              " tca@gnu.org, and he will tell you what the problem is." % filesystem.rcfile().encode("ascii", "backslashreplace"))), file=sys.stderr)
        print(file=sys.stderr)
        sys.exit("I give up (solfege.py)")
    except cfg.CfgParseException as e:
        i18n.setup(".")
        a, b = os.path.split(user_filename)
        renamed_fn = os.path.join(a, "BAD-" + b)
        m = Gtk.MessageDialog(None, Gtk.DialogFlags.MODAL, Gtk.MessageType.ERROR,
                Gtk.ButtonsType.NONE,
                _("Parsing %s failed") % filesystem.rcfile())
        m.format_secondary_text(str(e) + "\n\n" + _("We cannot recover from this, we can rename the corrupt file to %s and then start the program." % renamed_fn))
        m.add_buttons("Rename", 10)
        m.add_buttons(Gtk.STOCK_QUIT, 11)
        m.set_default_response(11)
        ret = m.run()
        if ret == 10:
            os.rename(user_filename, renamed_fn)
            m.destroy()
            cfg.initialise(app_defaults_filename, system_filename, user_filename)
        else:
            sys.exit(1)
    # MIGRATION from 2.9.2 to 2.9.3
    if cfg.get_string("app/lc_messages") == 'C (english)':
        cfg.set_string("app/lc_messages", "C")



# GNU Solfege - free ear training software
# vim: set fileencoding=utf-8 :
# Copyright (C) 2012, 2016 Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
sys.path.append(".")

import re

import solfege.parsetree as pt
from solfege.dataparser import Question


class LfMod(object):

    def __init__(self, builtins=None):
        if not builtins:
            builtins = {}
        self.m_builtins = builtins
        self.m_globals = builtins.copy()
        self.m_blocklists = {}

    def dump(self):
        import pprint
        print("Globals:")
        pprint.pprint(self.m_globals)
        print("Blocks:")
        pprint.pprint(self.m_blocklists)
        print("--------")

translation_re = re.compile("(?P<varname>\w+)\[(?P<lang>[\w_+]+)\]")


def do_assignment(mod, statement, local_namespace, global_namespace, in_header,
        included):
    """
    in_header is True if the assignment is done inside a header block. We
    need to treat assignments to the 'module' variable special, since we
    must handle module names that are equal to already defined variables
    and functions.
    """
    if isinstance(statement.right, pt.Program):
        local_namespace[str(statement.left)] = parse_tree_interpreter(statement.right, mod.m_builtins)
    else:
        m = translation_re.match(str(statement.left))
        if m:
            if m.group('varname') not in local_namespace:
                print("FIXME: correct exception aka LessonfileException")
                raise Exception("Define the C-locale variable before adding translations")
            if in_header and included:
                if m.group('varname') in local_namespace:
                    return
            local_namespace[m.group('varname')] = \
                local_namespace[m.group('varname')].add_translation(
                    m.group('lang'),
                    statement.right.evaluate(local_namespace, global_namespace))
        else:
            if in_header:
                if included:
                    if str(statement.left) in local_namespace:
                        return
                if in_header and statement.left == 'module':
                    local_namespace[str(statement.left)] = str(statement.right)
                    return
            local_namespace[str(statement.left)] = statement.right.evaluate(local_namespace, global_namespace)


def do_module(block, mod, included=False):
    assert isinstance(mod, LfMod)
    assert isinstance(block, pt.Program)
    for statement in block:
        if isinstance(statement, pt.Assignment):
            # On the module top level, the local namespace it the
            # global namespace
            try:
                do_assignment(mod, statement, mod.m_globals, mod.m_globals, False, included)
            except pt.ParseTreeException as e:
                e.m_nonwrapped_text = block._lexer.get_err_context(e.m_tokenpos)
                raise
        elif isinstance(statement, (pt.Block, pt.NamedBlock)):
            blocks = mod.m_blocklists.setdefault(statement.m_blocktype, [])
            if statement.m_blocktype == 'question':
                blocks.append(Question())
            elif statement.m_blocktype == 'header':
                if not blocks:
                    blocks.append({})
            else:
                blocks.append({})
            if isinstance(statement, pt.NamedBlock):
                mod.m_globals[statement.m_name] = blocks[-1]
                # Named blocks need to know their name. It is not enough
                # that their name point to them from the global name space.
                # Evercises like elembuilder need to lookup their name
                # like a dict: elem['name']
                blocks[-1]['name'] = statement.m_name
            for block_statement in statement:
                try:
                    do_assignment(mod, block_statement, blocks[-1], mod.m_globals,
                              statement.m_blocktype == 'header', included)
                except pt.ParseTreeException as e:
                    e.m_nonwrapped_text = block._lexer.get_err_context(e.m_tokenpos)
                    raise
        elif isinstance(statement, pt.IncludeStatement):
            do_module(statement.m_inctree, mod, included=True)
    return mod


def parse_tree_interpreter(tree, builtins=None):
    """
    Interpret a parse tree from solfege.parsetree into LfMod objects
    """
    mod = LfMod(builtins)
    do_module(tree, mod)
    return mod



# vim: set fileencoding=utf-8 :
# GNU Solfege - free ear training software
# Copyright (C) 2010, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
import urllib.request, urllib.parse, urllib.error

from gi.repository import Gtk

import solfege
from solfege import gu

pyalsa_ver = "1.0.26"

try:
    import pyalsa
except ImportError:
    pyalsa = None


def download():
    pdir = os.path.expanduser("~")
    url = "ftp://ftp.alsa-project.org/pub/pyalsa/pyalsa-%s.tar.bz2" % pyalsa_ver
    bz2 = "pyalsa-%s.tar.bz2" % pyalsa_ver
    _pyalsa_ver = pyalsa_ver
    bz2abs = os.path.join(pdir, bz2)

    m = Gtk.MessageDialog(solfege.win, Gtk.DialogFlags.MODAL, Gtk.MessageType.QUESTION,
            message_format="Download python modules?")
    m.add_button("Cancel", Gtk.ResponseType.CANCEL)
    m.add_button(Gtk.STOCK_EXECUTE, Gtk.ResponseType.ACCEPT)
    m.format_secondary_markup("""This will download «%(url)s» and build it in a subdirectory of %(pdir)s.

This is what the program will do for you:

<span font_family="monospace">$ wget %(url)s
$ tar xjf %(bz2)s
$ cd pyalsa-%(_pyalsa_ver)s
$ python setup.py build</span>

You will be given instructions how to make Solfege find the module after
it have been built.

If you get an error message about missing alsa/asoundlib.h, then you
must install the ALSA kernel headers. On Debian or Ubuntu you do this
by installing the libasound2-dev package.
____________________________________________________________________________________________________
""" % locals())
    # The long ____ in the string above is there to make the dialog wider
    # so that the command lines does not wrap.
    m.set_default_response(Gtk.ResponseType.YES)
    ret = m.run()
    m.destroy()
    if ret in (Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT):
        return
    logwin = gu.LogWindow(solfege.win)

    def progress_callback(count, size, total):
        if total == -1:
            total = 'unknown'
        logwin.write("Downloading %s of %s bytes\n" % (count * size, total))
    try:
        if not os.path.exists(bz2abs):
            urllib.request.urlretrieve(url, bz2abs, progress_callback)
    except IOError as e:
        logwin.write(str(e).decode(sys.getfilesystemencoding(), 'replace'))
        logwin.write("\nFailed to download alsa Python modules.")
        logwin.run_finished()
        return
    try:
        logwin.popen(["tar", "xjf", bz2], cwd=pdir)
    except OSError as e:
        logwin.write("\nExtracting %s failed.\n" % bz2)
        logwin.write("Make sure tar and bz2 is installed.\n")
        logwin.write("Could not build ALSA python module.\n")
        logwin.run_finished()
        return
    try:
        ret = logwin.popen(["python", "setup.py", "build"],
            cwd=os.path.join(pdir, "pyalsa-%s" % pyalsa_ver))
        if ret != 0:
            logwin.write("\nRunning the python interpreter failed.\nCould not build ALSA python module.")
            logwin.run_finished()
            return
    except OSError as e:
        logwin.write("\nRunning the python interpreter failed.\nCould not build ALSA python module.")
        logwin.run_finished()
        return
    sys.path.append(os.path.join(pdir, "pyalsa-%s" % pyalsa_ver))
    import pyalsa
    reload(pyalsa)
    print(pyalsa)
    logwin.write("\npyalsa module: %s\n\n" % str(pyalsa))
    logwin.write("The module is built. Now you must make Solfege find it.\n")
    logwin.write("There are some ways to do it:\n")
    logwin.write("\nIf you run bash, you can add this to ~/.bashrc:\n")
    logwin.write("export PYTHONPATH=%s\n\n" % os.path.join(pdir, "pyalsa-%s" % pyalsa_ver))
    logwin.write("Or start solfege this way:\n")
    logwin.write("$ PYTHONPATH=%s solfege\n" % os.path.join(pdir, "pyalsa-%s" % pyalsa_ver))
    logwin.write("\nOr install it so that it is found automatically:")
    logwin.write(("\n$ cd %(pdir)s/pyalsa-" % locals()) + pyalsa_ver)
    logwin.write("\n$ sudo python setup.py install")
    logwin.write("\n\nOr if you don't have sudo setup:")
    logwin.write('\n$ su -c "python setup.py install"')
    logwin.write("")

    logwin.run_finished()



# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING


import unittest
import os

from solfege.testlib import outdir
from solfege.reportlib import *


class TestReport(unittest.TestCase):

    def create_report(self):
        r = Report()
        r.append(Heading(1, "Heading1"))
        r.append(Paragraph("Text"))
        t = Table()
        r.append(t)
        row = TableRow()
        row.append("Cell1 text")
        row.append("Cell2 text")
        t.append(row)
        t.append_row("cell1", "cell2")
        return r

    def test_html(self):
        r = self.create_report()
        HtmlReport(r, os.path.join(outdir, "t1.html"))
        os.remove(os.path.join(outdir, "t1.html"))

    def test_latex(self):
        r = self.create_report()
        LatexReport(r, os.path.join(outdir, "t1.tex"))
        os.remove(os.path.join(outdir, "t1.tex"))

suite = unittest.makeSuite(TestReport)



# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2010, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING

import unittest
from solfege.mpd.musicalpitch import MusicalPitch, InvalidNotenameException
from solfege.mpd.elems import (Note, Voice, Score, BarProxy, RhythmStaff,
                Rest, TimeSignature, Bar, Skip, Stem)
from solfege.mpd.rat import Rat
from solfege.mpd.duration import Duration
from solfege.mpd import const
from solfege.mpd import parser
from solfege import mpd


def f3(s):
    return parser.parse_to_score_object(s).get_timelist()


class TestScore(unittest.TestCase):

    def setUp(self):
        self.score = Score()
        self.staff = self.score.add_staff()  # default class is Staff
        self.__bp = None

    def get_bp(self):
        # We need a getter for .bp because we cannot create the BarProxy
        # before some elements have been added to the bar.
        if not self.__bp:
            self.__bp = BarProxy(self.score.voice11, Rat(0, 1))
        return self.__bp
    bp = property(get_bp)

    def test_constructor_ok(self):
        # add_staff automatically creates the first voice
        self.assertEqual(len(self.staff.m_voices), 1)
        # and creates the voice shortcut
        self.assertEqual(type(self.score.voice11), Voice)

    def test_add_note(self):
        self.score.voice11.append(Note.new_from_string("c'4"))
        self.score.voice11.append(Note.new_from_string("c'2"))
        self.score.voice11.append(Note.new_from_string("c'8"))
        self.assertIsInstance(self.score.voice11.m_tdict[Rat(0, 1)]['elem'][0], Note)
        self.assertIsInstance(self.score.voice11.m_tdict[Rat(0, 1)]['elem'][0].w_parent(), Stem)
        self.assertIsInstance(self.score.voice11.m_tdict[Rat(0, 1)]['elem'][0].w_parent().w_parent(), Voice)
        # bar full:
        self.assertRaises(Voice.BarFullException, self.score.voice11.append,
            Note.new_from_string("c'4"))
        self.score.voice11.append(Note.new_from_string("c'8"))
        self.score.voice11.append([
            Note.new_from_string("c'1"),
            Note.new_from_string("e'1"),
            Note.new_from_string("g'1"),
        ])

    def test_add_rest(self):
        self.score.voice11.append(Rest(Duration.new_from_string("8")))
        self.assertIsInstance(self.score.voice11.m_tdict[Rat(0, 1)]['elem'][0], Rest)
        self.assertIsInstance(self.score.voice11.m_tdict[Rat(0, 1)]['elem'][0].w_parent(), Voice)

    def test_add_bar(self):
        self.assertIsInstance(self.score.add_bar(TimeSignature(4, 4)), Bar)
        self.score.add_bar(TimeSignature(4, 4))
        self.score.add_bar(TimeSignature(1, 4))
        self.score.add_bar(TimeSignature(1, 4))
        self.assertEqual(self.score.m_bars[0].m_timepos, Rat(0, 1))
        self.assertEqual(self.score.m_bars[1].m_timepos, Rat(1, 1))
        self.assertEqual(self.score.m_bars[2].m_timepos, Rat(2, 1))
        self.assertEqual(self.score.m_bars[3].m_timepos, Rat(9, 4))

    def test_add_bar_autotimesig(self):
        self.score.add_bar(None)
        self.assertEqual(self.score.m_bars[0].m_timesig, TimeSignature(4, 4))
        self.score.add_bar(TimeSignature(5, 8))
        self.assertEqual(self.score.m_bars[1].m_timesig, TimeSignature(5, 8))
        self.score.add_bar(None)
        self.assertEqual(self.score.m_bars[2].m_timesig, TimeSignature(5, 8))

    def test_add_partial_bar(self):
        self.score.add_partial_bar(Duration.new_from_string("4"), TimeSignature(4, 4))
        self.score.add_bar(None)
        self.assertEqual(self.score.m_bars[0].m_timepos, Rat(0, 1))
        self.assertEqual(self.score.m_bars[0].end(), Rat(1, 4))
        self.assertEqual(self.score.m_bars[1].m_timepos, Rat(1, 4))
        self.assertEqual(self.score.get_bar_at(Rat(0, 1)), self.score.m_bars[0])
        self.assertEqual(self.score.get_bar_at(Rat(1, 4)), self.score.m_bars[1])
        self.score.voice11.append(Note.new_from_string("c4"))
        self.score.voice11.append(Note.new_from_string("c1"))
        self.score.voice11.append(Note.new_from_string("c1"))

    def test_get_bar_at(self):
        # Raise exception, since we have not added bars
        self.assertRaises(IndexError, self.score.get_bar_at, Rat(0, 1))
        # None is to get the default timesig 4/4
        self.score.add_bar(None)
        self.assertIsInstance(self.score.get_bar_at(Rat(0, 1)), Bar)

    def test_set_clef(self):
        self.score.voice11.set_clef("violin")
        self.assertEqual(self.score.staff1.m_tdict[Rat(0, 1)]['clef'].m_name, "violin")
        self.score.voice11.set_clef("bass")
        # The last clef set is remembered
        self.assertEqual(self.score.staff1.m_tdict[Rat(0, 1)]['clef'].m_name, "bass")
        self.score.voice11.append(Note.new_from_string("c'8"))
        self.score.voice11.set_clef("treble")
        self.assertEqual(self.score.staff1.m_tdict[Rat(0, 1)]['clef'].m_name, "bass")
        self.assertEqual(self.score.staff1.m_tdict[Rat(1, 8)]['clef'].m_name, "treble")

    def test_stem_down(self):
        self.score.voice11.append(Note.new_from_string("c'8"))
        self.score.voice11.append(Note.new_from_string("c'8"), const.DOWN)
        self.assertEqual(self.score.voice11.m_tdict[Rat(0, 1)]['elem'].m_stemdir, const.BOTH)
        self.assertEqual(self.score.voice11.m_tdict[Rat(1, 8)]['elem'].m_stemdir, const.DOWN)

    def test_one_voice_in_rhythm_staff(self):
        staff = self.score.add_staff(RhythmStaff)
        self.assertRaises(staff.OnlyOneVoiceException,
            staff.add_voice)

    def test_no_chords_in_rhythm_staff(self):
        self.score.add_staff(RhythmStaff)
        self.score.voice21.append(Note.new_from_string("cis4"))
        self.assertRaises(Voice.NoChordsInRhythmStaffException,
                self.score.voice21.append,
                [Note.new_from_string("gis4"), Note.new_from_string("fis4")])

    def test_is_bar_full(self):
        self.assertTrue(self.score.voice11.is_bar_full())
        self.score.voice11.append(Note.new_from_string("c2"))
        self.assertFalse(self.score.voice11.is_bar_full())
        self.score.voice11.append(Note.new_from_string("c4"))
        self.assertFalse(self.score.voice11.is_bar_full())
        self.score.voice11.append(Note.new_from_string("c4"))
        self.assertTrue(self.score.voice11.is_bar_full())

    def test_midigen_rest(self):
        self.score.voice11.append(Note.new_from_string("c4"))
        self.score.voice11.append(Rest(Duration.new_from_string("4")))
        self.score.voice11.append(Note.new_from_string("c4"))
        t = mpd.score_to_tracks(self.score)
        self.assertEqual(t[0].str_repr(), "n48 d1/4 o48 d1/4 n48 d1/4 o48")

    def test_midigen_tie(self):
        n1 = Note.new_from_string("c4")
        n2 = Note.new_from_string("c4")
        n3 = Note.new_from_string("c4")
        n4 = Note.new_from_string("d4")
        self.score.voice11.append(n1)
        self.score.voice11.append(n2)
        self.score.voice11.append(n3)
        self.score.voice11.append(n4)
        self.score.voice11.tie([n1, n2, n3])
        t = mpd.score_to_tracks(self.score)
        self.assertEqual(t[0].str_repr(), "n48 d3/4 o48 n50 d1/4 o50")

    def test_tuplets(self):
        n1 = Note(MusicalPitch.new_from_notename("g'"),
                  Duration(8, 0, Rat(2, 3)))
        n2 = Note(MusicalPitch.new_from_notename("g'"),
                  Duration(8, 0, Rat(2, 3)))
        n3 = Note(MusicalPitch.new_from_notename("g'"),
                  Duration(8, 0, Rat(2, 3)))
        self.score.voice11.append(n1)
        self.score.voice11.append(n2)
        self.score.voice11.append(n3)
        self.score.voice11.tuplet(Rat(2, 3), const.UP, [n1, n2, n3])

    def test_single_tuplet(self):
        n1 = Note(MusicalPitch.new_from_notename("g'"),
                  Duration(8, 0, Rat(2, 3)))
        self.score.voice11.append(n1)
        self.score.voice11.tuplet(Rat(2, 3), const.UP, [n1])
        self.assertEqual(n1.w_parent().m_tupletinfo, 'end')

    def test_rh_1(self):
        self.assertEqual(f3(r"\staff{c}"), [[True, Rat(1, 4)]])
        self.assertEqual(f3(r"\staff{c2}"), [[True, Rat(1, 2)]])
        self.assertEqual(f3(r"\staff{c4 c8}"), [
            [True, Rat(1, 4)],
            [True, Rat(1, 8)],
        ])
        self.assertEqual(f3(r"\staff{c4 c4}"), [
            [True, Rat(1, 4)],
            [True, Rat(1, 4)],
        ])
        self.assertEqual(f3(r"\staff{c4. c4}"), [
            [True, Rat(3, 8)],
            [True, Rat(1, 4)],
        ])
        self.assertEqual(f3(r"\staff{c4~ c8 c4}"), [
            [True, Rat(3, 8)],
            [True, Rat(1, 4)],
        ])

    def test_rh_rest(self):
        self.assertEqual(f3(r"\staff{c4 r8 c4}"), [
            [True, Rat(1, 4)],
            [False, Rat(1, 8)],
            [True, Rat(1, 4)],
        ])
        self.assertEqual(f3(r"\staff{c4 r8 r16 c4}"), [
            [True, Rat(1, 4)],
            [False, Rat(3, 16)],
            [True, Rat(1, 4)],
        ])
        self.assertEqual(f3(r"\staff{c4 c8 c8}"), [
            [True, Rat(1, 4)],
            [True, Rat(1, 8)],
            [True, Rat(1, 8)],
        ])

    def test_rh_tie(self):
        self.assertEqual(f3(r"\staff{c4~ c8 c8 r4 c4}"), [
            [True, Rat(3, 8)],
            [True, Rat(1, 8)],
            [False, Rat(1, 4)],
            [True, Rat(1, 4)]
        ])

    def test_rh_1a(self):
        self.assertEqual(f3(r"\staff{c8~ c8~ c8}"), [
            [True, Rat(3, 8)],
        ])

    def test_rh_1b(self):
        self.assertEqual(f3(r"\staff{c4~ c8}"), [
            [True, Rat(3, 8)],
        ])

    def test_rh_2a(self):
        self.assertEqual(f3(r"\staff{c2}\addvoice{c4 c4}"), [
            [True, Rat(1, 4)],
            [True, Rat(1, 4)],
        ])

    def test_rh_2b(self):
        self.assertEqual(f3(r"\staff{c2}\addvoice{c4}"), [
            [True, Rat(1, 2)],
        ])

    def test_rh_2c(self):
        self.assertEqual(f3(r"\staff{c2}\addvoice{r4 c4}"), [
            [True, Rat(1, 4)],
            [True, Rat(1, 4)],
        ])

    def test_rh_2d(self):
        self.assertEqual(f3(r"\staff{c2}\addvoice{r4 r4}"), [
            [True, Rat(1, 2)],
        ])

    def test_rh_2e(self):
        self.assertEqual(f3(r"\staff{c1}\addvoice{r4 g2 r4}"), [
            [True, Rat(1, 4)],
            [True, Rat(3, 4)],
        ])

    def test_rh_3(self):
        self.assertEqual(f3(r"\staff{c4}"), [
            [True, Rat(1, 4)],
        ])
        self.assertEqual(f3(r"\staff{c4 c8}"), [
            [True, Rat(1, 4)],
            [True, Rat(1, 8)]])
        self.assertEqual(f3(r"\staff{c4 c8}"), [
            [True, Rat(1, 4)],
            [True, Rat(1, 8)]])
        self.assertEqual(f3(r"\staff{c4 c8 c8}"), [
            [True, Rat(1, 4)],
            [True, Rat(1, 8)],
            [True, Rat(1, 8)]])
        self.assertEqual(f3(r"""\staff\relative c{
c4 c8 c8
}"""), [
            [True, Rat(1, 4)],
            [True, Rat(1, 8)],
            [True, Rat(1, 8)]])

    def test_bar_fill_skips(self):
        n1 = Note(MusicalPitch.new_from_notename("g'"),
                  Duration(4, 0))
        self.score.voice11.append(n1)
        self.score.voice11.append(Rest(Duration.new_from_string("4")))
        self.bp.fill_skips()
        self.assertTrue(isinstance(self.score.voice11.m_tdict[Rat(0, 1)]['elem'][0], Note))
        self.assertTrue(isinstance(self.score.voice11.m_tdict[Rat(1, 4)]['elem'][0], Rest))
        self.assertTrue(isinstance(self.score.voice11.m_tdict[Rat(1, 2)]['elem'][0], Skip))
        self.assertTrue(isinstance(self.score.voice11.m_tdict[Rat(3, 4)]['elem'][0], Skip))

    def test_repack_bar_1(self):
        """
        We delete one note, and sees that the bar is packed correctly.
        """
        self.score.voice11.append(Note.new_from_string("c4"))
        self.score.voice11.append(Note.new_from_string("d4"))
        self.score.voice11.append(Note.new_from_string("e4"))
        self.score.voice11.append(Note.new_from_string("d4"))
        self.assertEqual(sorted(self.score.voice11.m_tdict.keys()),
            [Rat(0, 1), Rat(1, 4), Rat(1, 2), Rat(3, 4)])
        del self.score.voice11.m_tdict[Rat(1, 2)]
        self.assertEqual(sorted(self.score.voice11.m_tdict.keys()),
            [Rat(0, 1), Rat(1, 4), Rat(3, 4)])
        self.bp.repack()
        self.assertEqual(sorted(self.score.voice11.m_tdict.keys()),
            [Rat(0, 1), Rat(1, 4), Rat(1, 2)])

    def test_repack_bar_2(self):
        """
        One note is changed from 1/4 to 1/2 length. Then we call
        bar.repack(voice11) and we see that the last note is moved
        further out in the bar. This works ok since the bar was not
        full. There where not Skips at the end of the bar.
        """
        self.score.voice11.append(Note.new_from_string("c4"))
        self.score.voice11.append(Note.new_from_string("d4"))
        self.score.voice11.append(Note.new_from_string("e4"))
        self.assertEqual(sorted(self.score.voice11.m_tdict.keys()),
            [Rat(0, 1), Rat(1, 4), Rat(1, 2)])
        self.score.voice11.m_tdict[Rat(1, 4)]['elem'][0] = Note.new_from_string("g2")
        self.assertEqual(sorted(self.score.voice11.m_tdict.keys()),
            [Rat(0, 1), Rat(1, 4), Rat(1, 2)])
        self.bp.repack()
        self.assertEqual(sorted(self.score.voice11.m_tdict.keys()),
            [Rat(0, 1), Rat(1, 4), Rat(3, 4)])

    def test_repack_bar_3(self):
        """
        One note is changed from 1/4 to 1/1 length. Then we call
        bar.repack and a BarFullException is raised.
        """
        self.score.voice11.append(Note.new_from_string("c4"))
        self.score.voice11.append(Note.new_from_string("d4"))
        self.score.voice11.append(Note.new_from_string("e4"))
        self.assertEqual(sorted(self.score.voice11.m_tdict.keys()),
            [Rat(0, 1), Rat(1, 4), Rat(1, 2)])
        self.score.voice11.m_tdict[Rat(1, 4)]['elem'][0] = Note.new_from_string("g1")
        self.assertEqual(sorted(self.score.voice11.m_tdict.keys()),
            [Rat(0, 1), Rat(1, 4), Rat(1, 2)])
        self.assertRaises(Voice.BarFullException, self.bp.repack)

    def test_get_timeposes_of(self):
        self.score.add_bar(TimeSignature(4, 4))
        self.score.add_bar(TimeSignature(4, 4))
        self.assertEqual(
            self.score.voice11.get_timeposes_of(self.score.m_bars[0]), [])
        self.assertEqual(
            self.score.voice11.get_timeposes_of(self.score.m_bars[1]), [])
        self.score.voice11.append(Note.new_from_string("c4"))
        self.assertEqual(
            self.score.voice11.get_timeposes_of(self.score.m_bars[0]),
            [Rat(0, 1)])
        self.assertEqual(
            self.score.voice11.get_timeposes_of(self.score.m_bars[1]), [])

    def test_get_time_pitch_list(self):
        def f(s, tempo=(60, 4)):
            s = parser.parse_to_score_object(r"\staff{ %s}" % s)
            return s.voice11.get_time_pitch_list(tempo)
        self.assertEqual(f("c4 d4"), [(48, 1.0), (50, 1.0)])
        self.assertEqual(f("c4 d2"), [(48, 1.0), (50, 2.0)])
        self.assertEqual(f("c4 d2", (120, 4)), [(48, 0.5), (50, 1.0)])
        self.assertEqual(f(r" \times 2/3 { c4 d2 }"), [(48, 2.0 / 3), (50, 4.0 / 3)])
        self.assertEqual(f("c4 r8. d4"), [(48, 1.0), (-1, 0.75), (50, 1.0)])
        self.assertRaises(Voice.NotUnisonException, f, "<c4 d4>")

    def test_is_last(self):
        voice = self.score.voice11
        voice.append(Note.new_from_string("c2"))
        voice.append(Note.new_from_string("d2"))
        voice.append(Note.new_from_string("f1"))
        voice.append(Note.new_from_string("g4"))
        self.assertTrue(voice.is_last(Rat(1, 2)))
        self.assertFalse(voice.is_last(Rat(0, 1)))
        self.assertTrue(voice.is_last(Rat(1, 2)))
        self.assertTrue(voice.is_last(Rat(1, 1)))
        self.assertFalse(voice.is_last(Rat(2, 1)))


class TestNote(unittest.TestCase):

    def setUp(self):
        self.score = Score()
        self.staff = self.score.add_staff()  # default class is Staff

    def test_contructor(self):
        Note(MusicalPitch.new_from_notename("c'"),
             Duration.new_from_string("4."))
        self.assertRaises(AssertionError,
                          Note, "4.", MusicalPitch.new_from_notename("c'"))
        self.assertRaises(AssertionError,
                          Note, Duration.new_from_string("4."), "c'")

    def test_new_from_string(self):
        Note.new_from_string("c'4.")
        self.assertRaises(InvalidNotenameException, Note.new_from_string, "x")

    def test_beam(self):
        voice = self.score.voice11
        n1 = Note.new_from_string("c'8")
        n2 = Note.new_from_string("d'8")
        voice.append(n1)
        voice.append(n2)
        voice.beam([n1, n2])
        self.score.staff1.add_voice()
        n3 = Note.new_from_string("d'8")
        n4 = Note.new_from_string("d'8")
        self.assertRaises(Voice.NoteDontBelongHere, voice.beam, [n3, n4])

    def test_tie_2notes(self):
        voice = self.score.voice11
        n1 = Note.new_from_string("c'4")
        n2 = Note.new_from_string("c'8")
        voice.append(n1)
        voice.append(n2)
        voice.tie([n1, n2])

    def test_tie_3notes(self):
        voice = self.score.voice11
        n1 = Note.new_from_string("c'4")
        n2 = Note.new_from_string("c'4")
        n3 = Note.new_from_string("c'4")
        voice.append(n1)
        voice.append(n2)
        voice.append(n3)
        voice.tie([n1, n2, n3])


class TestBarProxy(unittest.TestCase):

    def setUp(self):
        self.score = Score()
        self.staff = self.score.add_staff()
        v = self.score.voice11
        v.append(Note.new_from_string("c'4"))
        v.append(Note.new_from_string("d'2"))
        v.append(Note.new_from_string("c'4"))
        v.append(Note.new_from_string("c'4"))
        v.append(Note.new_from_string("c'4"))
        v.append(Note.new_from_string("c'4"))
        v.append(Note.new_from_string("c'4"))

    def test_1(self):
        bp = BarProxy(self.score.voice11, Rat(0, 1))
        bp.remove_trailing(Rat(1, 4))
        bp.pop_last_elem()
        bp.remove_skips()
        bp.repack()
        bp.fill_skips()
        bp.end()

suite = unittest.makeSuite(TestScore)
suite.addTest(unittest.makeSuite(TestNote))
suite.addTest(unittest.makeSuite(TestBarProxy))



# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING

# Utility functions used by the test suite.


import builtins
import os
import unittest

from solfege import i18n

builtins.testsuite_is_running = True

outdir = 'test-outdir'


class I18nSetup(unittest.TestCase):

    def setUp(self):
        self.__saved_LANGUAGE = os.environ.get('LANGUAGE', None)
        os.environ['LANGUAGE'] = 'nb'
        i18n.setup(".")

    def tearDown(self):
        if self.__saved_LANGUAGE:
            os.environ['LANGUAGE'] = self.__saved_LANGUAGE


class TmpFileBase(unittest.TestCase):
    """
    Sub-classes must set .parserclass
    """
    tmpdir = "solfege/tests/tmp-lesson-files"

    def setUp(self):
        self.p = self.parserclass()
        self.m_files = set()

    def add_file(self, content, filename):
        self.m_files.add(filename)
        outfile = open(os.path.join(self.tmpdir, filename), 'w')
        outfile.write(content)
        outfile.close()

    def do_file(self, content):
        self.add_file(content, 'testfile')
        self.p.parse_file(os.path.join(self.tmpdir, 'testfile'))
        return self.p

    def tearDown(self):
        for filename in self.m_files:
            os.remove(os.path.join(self.tmpdir, filename))



# GNU Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import codecs


class Heading(object):

    def __init__(self, level, text):
        """
        1 is top level heading, 2 is below that.
        """
        self.m_level = level
        self.m_text = text


class Report(list):
    pass


class Paragraph(str):
    pass


class Table(list):

    def append_row(self, *cells):
        self.append(cells)


class TableRow(list):
    pass


class ReportWriterCommon(object):

    def __init__(self, report, filename):
        self.m_outfile = codecs.open(filename, "w", "utf-8")
        self.write_report(report)
        self.m_outfile.close()


class HtmlReport(ReportWriterCommon):

    def write_report(self, report):
        print("""<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8">
<style type="text/css">
th { text-align: left; border-bottom: 1px solid black}
</style>
</head>
<body>""", file=self.m_outfile)
        for elem in report:
            f = {'Heading': self.write_heading,
                 'Paragraph': self.write_paragraph,
                 'Table': self.write_table,
                }[elem.__class__.__name__](elem)
        print("</body>\n</html>", file=self.m_outfile)

    def write_heading(self, heading):
        print("<h%(level)i>%(str)s</h%(level)i>" % {
            'level': heading.m_level,
            'str': heading.m_text}, file=self.m_outfile)

    def write_paragraph(self, paragraph):
        print("<p>%s</p>" % paragraph, file=self.m_outfile)

    def write_table(self, table):
        print("<table border='0'>", file=self.m_outfile)
        for row in table:
            self.write_tablerow(row)
        print("</table>", file=self.m_outfile)

    def write_tablerow(self, row):
        print("<tr>", file=self.m_outfile)
        for t in row:
            print("<td>%s</td>" % t, file=self.m_outfile)
        print("</tr>", file=self.m_outfile)


class LatexReport(ReportWriterCommon):

    def write_report(self, report):
        print(r"\documentclass{article}", file=self.m_outfile)
        print(r"\begin{document}", file=self.m_outfile)
        for elem in report:
            f = {'Heading': self.write_heading,
                 'Paragraph': self.write_paragraph,
                 'Table': self.write_table,
            }[elem.__class__.__name__](elem)
        print(r"\end{document}", file=self.m_outfile)

    def write_heading(self, heading):
        print(r"\section{%s}" % heading.m_text, file=self.m_outfile)

    def write_paragraph(self, paragraph):
        print(paragraph, file=self.m_outfile)

    def write_table(self, t):
        pass

    def write_tablerow(self, t):
        pass



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject
from gi.repository import Gtk

from solfege import abstract
from solfege import gu
from solfege import lessonfile
from solfege import mpd
from solfege import soundcard
from solfege import statistics, statisticsviewer
from solfege import utils

import solfege


class Teacher(abstract.Teacher):
    OK = 0
    ERR_PICKY = 1
    ERR_NO_INTERVALLS = 2
    ERR_NOTERANGE = 3

    def __init__(self, exname):
        abstract.Teacher.__init__(self, exname)
        self.lessonfileclass = lessonfile.IntervalsLessonfile
        self.m_statistics = statistics.HarmonicIntervalStatistics(self)
        self.m_tonika = None

    def enter_test_mode(self):
        self.m_custom_mode = False
        self.m_statistics.enter_test_mode()
        self.m_P.enter_test_mode()

    def new_question(self, L, H):
        self.q_status = self.QSTATUS_NEW
        return Teacher.OK

    def guess_answer(self, answer):
        """
        Return: 1 if correct, None if wrong
        """
        assert self.q_status not in (self.QSTATUS_NO, self.QSTATUS_GIVE_UP)
        if 1:  # add test
            # if self.q_status == self.QSTATUS_NEW \
            #        and not self.m_custom_mode:
            #    #self.m_statistics.add_correct(some value)
            self.maybe_auto_new_question()
            self.q_status = self.QSTATUS_SOLVED
            return 1
        else:
            if self.q_status == self.QSTATUS_NEW:
                # if not self.m_custom_mode:
                #    self.m_statistics.add_wrong(some value, some value)
                self.q_status = self.QSTATUS_WRONG
            if solfege.app.m_test_mode:
                self.maybe_auto_new_question()

    def give_up(self):
        """This function is only called *after* the user already has
        answered wrong once, so the statistics are already updated.
        """
        self.q_status = self.QSTATUS_GIVE_UP

    def play_question(self):
        return

    def start_practise(self):
        return


class Gui(abstract.Gui):
    lesson_heading = _("Identify the interval")

    def __init__(self, teacher):
        abstract.Gui.__init__(self, teacher)
        ################
        # practice_box #
        ################
        self.std_buttons_add(
            ('new', None),  # add callback
            ('give_up', self.give_up),
        )
        self.g_flashbar = gu.FlashBar()
        self.g_flashbar.show()
        self.practise_box.pack_start(self.g_flashbar, False, False, 0)
        self.practise_box.set_spacing(gu.PAD)
        ###############
        # config_grid #
        ###############
        self._add_auto_new_question_gui(row=0)
        # ----------------------------------------------
        self.config_grid.show_all()
        ##############
        # statistics #
        ##############
        self.setup_statisticsviewer(statisticsviewer.StatisticsViewer,
                                   _("Statistics for this example exercise"))

    def give_up(self, _o=None):
        if self.m_t.q_status == self.QSTATUS_WRONG:
            self.g_flashbar.push(_("The answer is: xxx"))
            self.m_t.give_up()
            self.std_buttons_give_up()

    def new_question(self, _o=None):
        """This function is called when you click on the 'New'
        button, or when you use the key bindings. So it can be called even
        if the 'New' button is insensitive.
        """
        self.std_buttons_new_question()
        print("New question")

    def on_start_practise(self):
        self.m_t.start_practise()
        super(Gui, self).on_start_practise()
        if not self.m_t.m_custom_mode:
            self.m_t.m_statistics.reset_session()
        self.g_statview.g_heading.set_text("%s - %s" % (_("Example"), self.m_t.m_P.header.title))
        self.std_buttons_start_practise()
        if solfege.app.m_test_mode:
            self.g_flashbar.delayed_flash(self.short_delay,
                _("Click 'Start test' to begin."))
        else:
            self.g_flashbar.delayed_flash(self.short_delay,
                _("Click 'New' to begin."))

    def on_end_practise(self):
        self.m_t.end_practise()
        self.g_flashbar.clear()

    def enter_test_mode(self):
        self.m_saved_q_auto = self.get_bool('new_question_automatically')
        self.m_saved_s_new = self.get_float('seconds_before_new_question')
        self.set_bool('new_question_automatically', True)
        self.set_float('seconds_before_new_question', 0.5)
        self.m_t.enter_test_mode()
        self.g_new.set_label(_("_Start test"))
        self.g_cancel_test.show()
        self.g_give_up.hide()

    def exit_test_mode(self):
        self.set_bool('new_question_automatically', self.m_saved_q_auto)
        self.set_float('seconds_before_new_question', self.m_saved_s_new)
        self.m_t.exit_test_mode()
        self.g_new.show()
        self.g_new.set_label(_("_New something"))
        self.g_give_up.show()



# GNU Solfege - free ear training software
# Copyright (C) 2010, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from solfege import gu
from solfege import abstract
from solfege import lessonfile
from solfege import mpd
from solfege import soundcard

from solfege.mpd import Duration
from solfege.mpd import elems
from solfege.mpd import RhythmWidget, RhythmWidgetController


class Teacher(abstract.Teacher):
    ERR_PICKY = 1
    OK = 2

    def __init__(self, exname):
        abstract.Teacher.__init__(self, exname)
        self.lessonfileclass = lessonfile.RhythmDictation2Lessonfile

    def new_question(self):
        """
        We will create a timelist of the question when we create the
        question, and compare it to the RhythmStaff.get_timelist.
        We will also create a  PercussionTrack that will be used when
        we play the question.
        """
        if self.get_bool('config/picky_on_new_question') \
                and self.q_status in [self.QSTATUS_NEW, self.QSTATUS_WRONG]:
            return self.ERR_PICKY
        self.q_status = self.QSTATUS_NEW
        self.m_P.generate_random_question()
        self.m_score = self.m_P.m_answer_score
        return self.OK

    def play_question(self):
        score = self.m_P.m_question_score
        countin = self.m_P.m_globals.get('countin')
        countin = self.m_P.m_questions[self.m_P._idx].get('countin', countin)
        if countin:
            score = elems.Score.concat2(countin.get_score(self.m_P, as_name='countin'), score)
        tracks = mpd.score_to_tracks(score)
        tracks[0].prepend_bpm(*self.m_P.get_tempo())
        soundcard.synth.play_track(*tracks)

    def guess_answer(self, staff):
        assert self.q_status not in (self.QSTATUS_NO, self.QSTATUS_GIVE_UP)
        if self.m_P.m_question_score.get_timelist() == self.m_P.m_answer_score.get_timelist():
            self.q_status = self.QSTATUS_SOLVED
            return True
        else:
            self.q_status = self.QSTATUS_WRONG
            return False

    def give_up(self):
        self.q_status = self.QSTATUS_GIVE_UP


class Gui(abstract.LessonbasedGui):

    def __init__(self, teacher):
        abstract.LessonbasedGui.__init__(self, teacher)
        self.g_w = RhythmWidget()
        self.g_w.connect('score-updated', self.on_score_updated)
        self.practise_box.pack_start(self.g_w, False, False, 0)
        self.g_c = RhythmWidgetController(self.g_w)
        self.practise_box.pack_start(self.g_c, False, False, 0)
        self.g_flashbar = gu.FlashBar()
        self.practise_box.pack_start(self.g_flashbar, False, False, 0)
        self.g_flashbar.show()
        self.std_buttons_add(
            ('new', self.new_question),
            ('guess_answer', self.guess_answer),
            ('repeat', self.repeat_question),
            ('give_up', self.give_up))
        self.g_w.show()

    def on_score_updated(self, w):
        self.g_guess_answer.set_sensitive(bool(self.g_w.m_score.get_timelist()))

    def new_question(self, *w):
        def exception_cleanup():
            self.m_t.q_status = self.m_t.QSTATUS_NO
            self.std_buttons_exception_cleanup()
        try:
            g = self.m_t.new_question()
            if g == self.m_t.OK:
                self.m_t.play_question()
                self.std_buttons_new_question()
                self.g_w.grab_focus()
                self.g_w.set_score(self.m_t.m_score)
                self.g_c.set_editable(True)
        except Duration.BadStringException as e:
            gu.dialog_ok(
                "Lesson file error",
                solfege.win,
                secondary_text="Bad rhythm string in the elements variable of the lessonfile. Only digits and dots expected: %s" % str(e))
            exception_cleanup()
        except Exception as e:
            if not self.standard_exception_handler(e, exception_cleanup):
                raise

    def guess_answer(self, *w):
        if self.m_t.q_status == Teacher.QSTATUS_SOLVED:
            if self.m_t.guess_answer(self.g_w.m_score):
                self.g_flashbar.flash(_("Correct, but you have already solved this question"))
            else:
                self.g_flashbar.flash(_("Wrong, but you have already solved this question"))
        else:
            if self.m_t.guess_answer(self.g_w.m_score):
                self.g_flashbar.flash(_("Correct"))
                self.std_buttons_answer_correct()
            else:
                self.g_flashbar.flash(_("Wrong"))
                self.std_buttons_answer_wrong()
                self.g_w.grab_focus()

    def repeat_question(self, *w):
        self.g_w.grab_focus()
        self.m_t.play_question()

    def give_up(self, *w):
        # Make a copy of the question asked. Then attach the staff
        # the user entered below, set some labels and display it.
        score_copy = self.m_t.m_P.m_question_score.copy()
        score_copy.m_staffs.append(self.g_w.m_score.m_staffs[0])
        score_copy.m_staffs[-1].set_parent(self.m_t.m_P.m_question_score)
        score_copy.m_staffs[0].m_label = _("The music played:")
        score_copy.m_staffs[-1].m_label = _("The rhythm you entered:")
        score_copy.create_shortcuts()
        self.g_w.set_score(score_copy, cursor=None)
        self.m_t.give_up()
        self.g_c.set_editable(False)
        self.std_buttons_give_up()

    def on_start_practise(self):
        super(Gui, self).on_start_practise()
        self.std_buttons_start_practise()
        self.g_c.set_editable(False)
        self.g_w.set_score(elems.Score())
        self.g_flashbar.delayed_flash(self.short_delay,
            _("Click 'New' to begin."))

    def on_end_practise(self):
        super(Gui, self).on_end_practise()
        self.g_w.set_score(elems.Score())



# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING

from __future__ import absolute_import
import unittest

import glob
import os.path
modules = [os.path.splitext(os.path.basename(x))[0]
           for x in glob.glob("solfege/soundcard/tests/test_*.py")]

for m in modules:
    exec("import solfege.soundcard.tests.%s" % m)
suite = unittest.TestSuite([globals()[m].suite for m in modules])



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2006, 2007, 2008, 2011, 2016   Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import copy

from solfege.mpd.rat import Rat
from solfege.mpd import mfutils
from solfege.mpd.const import DEFAULT_VELOCITY, DEFAULT_VOLUME

set_patch_delay = 0


class EventBase(object):

    def __init__(self):
        self.m_time = None

    def __str__(self):
        return "(%s, time:%s)" % (self.__class__.__name__, self.m_time)


class NoteEventBase(EventBase):

    def __init__(self, pitch, velocity):
        EventBase.__init__(self)
        assert 0 <= pitch
        self.m_pitch = pitch
        self.m_velocity = velocity

    def __str__(self):
        return "(%s, pitch:%s, vel:%s, time:%s)" % (self.__class__.__name__, self.m_pitch, self.m_velocity, self.m_time)


class NoteOnEvent(NoteEventBase):

    def __init__(self, pitch, velocity):
        NoteEventBase.__init__(self, pitch, velocity)


class NoteOffEvent(NoteEventBase):

    def __init__(self, pitch, velocity):
        NoteEventBase.__init__(self, pitch, velocity)


class Delay(EventBase):

    def __init__(self, duration):
        """
        duration is a Rat. Rat(1, 4) denotes a quarter-note.
        """
        EventBase.__init__(self)
        self.m_duration = duration

    def __str__(self):
        return "(%s, dur:%s, time:%s)" % (self.__class__.__name__, self.m_duration, self.m_time)


class SetPatchEvent(EventBase):

    def __init__(self, patch):
        EventBase.__init__(self)
        assert 0 <= patch < 128
        self.m_patch = patch

    def __str__(self):
        return "(%s, time:%s, patch:%i)" % (self.__class__.__name__, self.m_time, self.m_patch)


class SetVolumeEvent(EventBase):

    def __init__(self, volume):
        EventBase.__init__(self)
        assert 0 <= volume < 256
        self.m_volume = int(volume)

    def __str__(self):

        return "(%s, time:%s, volume:%i)" % (self.__class__.__name__, self.m_time, self.m_volume)


class TempoEvent(EventBase):

    def __init__(self, bpm, notelen):
        EventBase.__init__(self)
        assert isinstance(bpm, int)
        self.m_bpm = bpm
        self.m_notelen = notelen

    def __str__(self):
        return "(%s, time:%s, bpm/notelen: %s/%s)" % (self.__class__.__name__, self.m_time, self.m_bpm, self.m_notelen)


class MidiEventStream(object):
    TEMPO = 'tempo'
    VOLUME = 'volume'
    NOTE_ON = 'note-on'
    NOTE_OFF = 'note-off'
    NOTELEN_TIME = 'notelen-time'
    BENDER = 'bender'
    SET_PATCH = 'program-change'

    class ChannelDevice(object):
        """
        Bad name, but I don't have a better idea right now. This
        class will handle all 16 midi channels.
        """
        # We are zero-indexed, so this is MIDI channel 10
        percussion_MIDI_channel = 9

        class MidiChannel(object):

            def __init__(self, number):
                self.m_number = number
                self.m_tones = set()

            def start_tone(self, i):
                assert i not in self.m_tones
                self.m_tones.add(i)

            def stop_tone(self, i):
                assert i in self.m_tones
                self.m_tones.remove(i)

            def is_silent(self):
                """
                Return True if no tones are playing on this channel.
                """
                return not bool(self.m_tones)

        def __init__(self, nc=16):
            self.free_MIDI_channels = []
            if nc > self.percussion_MIDI_channel:
                for i in (list(range(self.percussion_MIDI_channel))
                          + list(range(self.percussion_MIDI_channel, 16))):
                    self.free_MIDI_channels.append(self.MidiChannel(i))
            else:
                for i in range(nc):
                    self.free_MIDI_channels.append(self.MidiChannel(i))
            # The dict key will be the (patch number, volume).
            # The value is a list of MidiChannel object that play with
            # the patch and volume the key say.
            self.allocated_MIDI_channels = {}
            # This dict maps from MIDI channel number to the actual
            # MidiChannel object
            self.int_to_channel_object = {}
            for channel in self.free_MIDI_channels:
                self.int_to_channel_object[channel.m_number] = channel

        def set_channel_data(self, channel, key, data):
            setattr(self.int_to_channel_object[channel], key, data)

        def get_channel_data(self, channel, key):
            return getattr(self.int_to_channel_object[channel], key, None)

        def require_channel(self, pitch, patch, volume):
            key = (patch, volume)
            if key in self.allocated_MIDI_channels:
                for channel in self.allocated_MIDI_channels[key]:
                    if pitch not in channel.m_tones:
                        return channel.m_number
                else:
                    # we have to alloc a new MIDI channel if the pitch is
                    # already playing
                    self.allocated_MIDI_channels[key].append(self.alloc_channel())
                    return self.allocated_MIDI_channels[key][-1].m_number
            else:
                self.allocated_MIDI_channels[key] = [self.alloc_channel()]
                return self.allocated_MIDI_channels[key][-1].m_number
            return -1

        def alloc_channel(self):
            """
            Return a unused MIDI channel. Search for silent channels first,
            allocate a new if none found.
            """
            if self.free_MIDI_channels:
                return self.free_MIDI_channels.pop(0)
            # First try to find an allocated channel that is silent.
            for key, channel_list in list(self.allocated_MIDI_channels.items()):
                for idx, channel in enumerate(channel_list):
                    if channel.is_silent():
                        ret = self.allocated_MIDI_channels[key].pop(idx)
                        if not self.allocated_MIDI_channels[key]:
                            del self.allocated_MIDI_channels[key]
                        return ret
            raise Exception("FIXME: handle running out of MIDI channels!")

        def get_channel_for_patch(self, patch, volume):
            """
            Return the MIDI channel number we should use to play a tone
            with this patch number. Raise KeyError if no channel is allocated
            yet.
            """
            return self.allocated_MIDI_channels[patch, volume].m_number

        def start_note(self, channel, pitch):
            self.int_to_channel_object[channel].start_tone(pitch)

        def stop_note(self, channel, pitch):
            self.int_to_channel_object[channel].stop_tone(pitch)

        def is_playing(self, channel, pitch):
            return pitch in self.int_to_channel_object[channel].m_tones

    def __init__(self, *tracks):
        # The number of MIDI channels are changed in the test suite to easier
        # test handling of too few midi channels.
        self.num_MIDI_channels = 16
        self.m_tracks = tracks
        for track in self.m_tracks:
            track.calculate_event_times()

    def _create_dict_of_track(self, track):
        """
        Return a dict of the track, where the key is a list with
        all events with the same m_time variable.
        The dict values are a dict with three keys:
        NoteOffEvents, OtherEvents, NoteOnEvents
        """
        retval = {}
        for event in track.m_v:
            retval[event.m_time] = retval.get(event.m_time, [])
            retval[event.m_time].append(event)
        for key in retval:
            retval[key] = {
                'NoteOffEvents': [x for x in retval[key] if isinstance(x, NoteOffEvent)],
                'OtherEvents': [x for x in retval[key] if not isinstance(x, (NoteOffEvent, NoteOnEvent))],
                'NoteOnEvents': [x for x in retval[key] if isinstance(x, NoteOnEvent)]}
        return retval

    def sorted_events(self):
        """
        This method will rearrange the midi events so that all events of the
        same type is after each other. We do this because we want to be
        able to do a little hack after the last SET_PATCH event, it multiple
        SET_PATCH happen at the same time.
        """
        data = {self.NOTELEN_TIME: [],
                self.NOTE_OFF: [],
                self.TEMPO: [],
                self.SET_PATCH: [],
                self.VOLUME: [],
                self.NOTE_ON: []
                }
        for e in self:
            if e[0] == self.NOTELEN_TIME:
                for k in (self.NOTE_OFF, self.TEMPO, self.SET_PATCH,
                          self.VOLUME, self.NOTE_ON):
                    for event in data[k]:
                        yield event
                    data[k] = []
                yield e
            else:
                data[e[0]].append(e)
        for event in data[self.NOTE_OFF]:
            yield event

    def __iter__(self):
        ret = []
        for e in self.__mkevents():
            if e[0] == 'program-change' and ret:
                i = len(ret)
                while (isinstance(ret[i - 1][1], Rat) or ret[i - 1][1] != e[1]) and i > 1:
                    i -= 1
                if i < len(ret):
                    while ret[i][0] == 'program-change' and i < len(ret) - 1 and ret[i][1] < e[1]:
                        i += 1
                ret.insert(i, e)
            else:
                ret.append(e)
        for e in ret:
            yield e

    def __mkevents(self):
        # tpos_set will know all the positions in time where anything happens
        # on any staff
        tpos_set = set()
        for track in self.m_tracks:
            tpos_set.update([x.m_time for x in track.m_v if x.m_time])
        tracks2 = [self._create_dict_of_track(track) for track in self.m_tracks]
        tpos_list = list(tpos_set)
        tpos_list.sort()
        # We use this variable to remember which instrument
        # we want the track to play.
        track_state = {}
        for x in range(len(self.m_tracks)):
            track_state[x] = {'volume-requested': DEFAULT_VOLUME,
                              'patch-requested': 0}
        tempo_request = None
        tempo_current = None
        # We use this list of dicts to know which MIDI channel the tones are
        # playing on. The key of the dicts are the integer value representing
        # the pitch, and the value is the MIDI channel the tone is playing on.
        track_notes = []
        for x in range(len(self.m_tracks)):
            track_notes.append({})
        ch_dev = self.ChannelDevice(self.num_MIDI_channels)
        last_pos = Rat(0, 1)
        for tpos in tpos_list:
            if tpos != last_pos:  # Just to not insert before the first events
                yield self.NOTELEN_TIME, tpos - last_pos
            for idx, track in enumerate(tracks2):
                if tpos in track:
                    for e in track[tpos]['NoteOffEvents']:
                        if e.m_pitch not in track_notes[idx]:
                            # This could happen if the user adds extra NoteOffEvents or adds one
                            # with the wrong pitch.
                            logging.debug("not stopping, not playing now: %s", e)
                            continue
                        chn = track_notes[idx][e.m_pitch]
                        del track_notes[idx][e.m_pitch]
                        assert ch_dev.is_playing(chn, e.m_pitch)
                        ch_dev.stop_note(chn, e.m_pitch)
                        yield self.NOTE_OFF, chn, e.m_pitch, e.m_velocity
            for idx, track in enumerate(tracks2):
                if tpos in track:
                    for e in track[tpos]['OtherEvents']:
                        if isinstance(e, SetPatchEvent):
                            track_state[idx]['patch-requested'] = e.m_patch
                        elif isinstance(e, SetVolumeEvent):
                            track_state[idx]['volume-requested'] = e.m_volume
                        elif isinstance(e, TempoEvent):
                            tempo_request = e
                        else:
                            logging.debug("MidiEventStream: NOT HANDLING EVENT: %s", e)
            for idx, track in enumerate(tracks2):
                if tpos in track:
                    for e in track[tpos]['NoteOnEvents']:
                        assert e.m_pitch not in track_notes[idx]
                        if tempo_request != tempo_current:
                            yield self.TEMPO, tempo_request.m_bpm, tempo_request.m_notelen
                            tempo_current = tempo_request
                        if isinstance(self.m_tracks[idx], PercussionTrack):
                            chn = ch_dev.percussion_MIDI_channel
                            if ch_dev.get_channel_data(chn, 'volume') != track_state[idx]['volume-requested']:
                                ch_dev.set_channel_data(chn, 'volume', track_state[idx]['volume-requested'])
                                yield self.VOLUME, chn, track_state[idx]['volume-requested']
                        else:
                            chn = ch_dev.require_channel(
                                e.m_pitch,
                                track_state[idx]['patch-requested'],
                                track_state[idx]['volume-requested'])
                            if ch_dev.get_channel_data(chn, 'patch') != track_state[idx]['patch-requested']:
                                ch_dev.set_channel_data(chn, 'patch', track_state[idx]['patch-requested'])
                                yield self.SET_PATCH, chn, track_state[idx]['patch-requested']
                            if ch_dev.get_channel_data(chn, 'volume') != track_state[idx]['volume-requested']:
                                ch_dev.set_channel_data(chn, 'volume', track_state[idx]['volume-requested'])
                                yield self.VOLUME, chn, track_state[idx]['volume-requested']
                        if ch_dev.is_playing(chn, e.m_pitch):
                            logging.debug("MidiEventStream: ignoring duplicate tone: %s", e)
                            continue
                        track_notes[idx][e.m_pitch] = chn
                        # ch_dev must know which tones are sounding on which
                        # MIDI channels, so it can handle the midi resources.
                        ch_dev.start_note(chn, e.m_pitch)
                        yield self.NOTE_ON, chn, e.m_pitch, e.m_velocity
            last_pos = tpos

    def create_midifile(self, filename, appendstreams=[]):
        """
        filename -- a string naming the file to write the generated midi file to.
                    Will overwrite a existing file.
        appendstrings -- a list of additional MidiEventStreams to append to
                    the midi file.
        """
        v = []
        notelen = 0
        set_patch_flag = False
        v += mfutils.mf_tempo(60 * 4 / 4)
        for stream in [self] + appendstreams:
            for e in stream.sorted_events():
                if e[0] == self.TEMPO:
                    v = v + mfutils.mf_tempo(e[1] * 4 / e[2])
                elif e[0] == self.NOTELEN_TIME:
                    notelen = e[1]
                elif e[0] == self.NOTE_ON:
                    if set_patch_flag:
                        if set_patch_delay:
                            v = v + mfutils.mf_note_off(set_patch_delay, 0, 0, 0)
                        set_patch_flag = False
                    v = v + mfutils.mf_note_on(int(96 * 4 * notelen), e[1], e[2], e[3])
                    notelen = 0
                elif e[0] == self.NOTE_OFF:
                    v = v + mfutils.mf_note_off(int(96 * 4 * notelen), e[1], e[2], e[3])
                    notelen = 0
                elif e[0] == self.SET_PATCH:
                    v = v + mfutils.mf_program_change(e[1], e[2])
                    set_patch_flag = True
                elif e[0] == self.VOLUME:
                    v = v + mfutils.mf_volume_change(e[1], e[2])
                elif e[0] == self.BENDER:
                    logging.debug("create_midifile: FIXME todo: seq_bender for play_with_drvmidi")
                    # m.seq_bender(DEV, e[1], e[2])
                else:
                    raise Exception("mpd.track: Corrupt track error")
        f = open(filename, "wb")
        mfutils.MThd(f)
        f.write(b"MTrk")
        mfutils.write_int32(f, len(v) + 4)
        v = v + mfutils.mf_end_of_track()
        mfutils.write_vect(f, v)
        f.close()

    def str_repr(self, details=0):
        v = []
        for e in self:
            if e[0] == self.TEMPO:
                v.append("t%s/%s" % (e[1], e[2]))
            elif e[0] == self.NOTE_ON:
                if e[1] == 9:
                    v.append("P%s" % e[2])
                else:
                    if details == 0:
                        v.append("n%s" % e[2])
                    elif details == 1:
                        v.append("n%s:%s" % (e[1], e[2]))
            elif e[0] == self.NOTE_OFF:
                v.append("o%s" % e[2])
            elif e[0] == self.SET_PATCH:
                v.append("p%i:%i" % (e[1], e[2]))
            elif e[0] == self.VOLUME:
                v.append("v%i:%i" % (e[1], e[2]))
            elif e[0] == self.NOTELEN_TIME:
                v.append("d%i/%i" % (e[1].m_num, e[1].m_den))
        return " ".join(v)


class Track:
    """
    A pitch is represented by an integer value 0-127.
    * There can only be one instance of a pitch sounding at the same time.
    * There can only be one instrument sounding at the same time.
    Right now there are no code that checks that this is true while
    adding notes.
    """

    def txtdump(self):
        for event in self.m_v:
            print(event)

    def str_repr(self):
        retval = []
        for e in self.m_v:
            if isinstance(e, SetPatchEvent):
                retval.append('p%i' % e.m_patch)
            elif isinstance(e, TempoEvent):
                retval.append('t%i/%i' % (e.m_bpm, e.m_notelen))
            elif isinstance(e, SetVolumeEvent):
                retval.append('v%i' % e.m_volume)
            elif isinstance(e, NoteOnEvent):
                retval.append('n%i' % e.m_pitch)
            elif isinstance(e, NoteOffEvent):
                retval.append('o%i' % e.m_pitch)
            elif isinstance(e, Delay):
                retval.append('d%i/%i' % (e.m_duration.m_num, e.m_duration.m_den))
        return " ".join(retval)

    def __init__(self, default_velocity=None):
        if default_velocity is None:
            self.m_default_velocity = DEFAULT_VELOCITY
        else:
            self.m_default_velocity = default_velocity
        self.m_v = []

    def start_note(self, pitch, vel=None):
        assert 0 <= int(pitch) < 128
        if vel is None:
            vel = self.m_default_velocity
        assert 0 <= vel < 128
        self.m_v.append(NoteOnEvent(int(pitch), int(vel)))

    def stop_note(self, pitch, vel=None):
        assert 0 <= int(pitch) < 128
        if vel is None:
            vel = self.m_default_velocity
        assert 0 <= vel < 128
        self.m_v.append(NoteOffEvent(int(pitch), int(vel)))

    def notelen_time(self, notelen):
        """
        To avoid having to alter all code calling this, we interpret
        notelen in two different ways depending on its type:
        int: replace to Rat(1, notelen)
        Rat: the value tell the note length. For example Rat(1, 4) for a
             quarter note.
        """
        if isinstance(notelen, int):
            self.m_v.append(Delay(Rat(1, notelen)))
        else:
            assert isinstance(notelen, Rat)
            self.m_v.append(Delay(notelen))

    def note(self, notelen, pitch, vel=None):
        """
        See notelen_time docstring.
        """
        if vel is None:
            vel = self.m_default_velocity
        assert 0 <= vel < 128
        self.start_note(pitch, vel)
        self.notelen_time(notelen)
        self.stop_note(pitch, vel)

    def set_patch(self, patch):
        """
        Add an event that will change the midi instrument for the
        notes following this event.
        """
        self.m_v.append(SetPatchEvent(patch))

    def prepend_patch(self, patch):
        """
        Insert an event that will change the midi instrument at the
        beginning of the track. If you call this method several times,
        only the first call will have any effect.
        """
        self.m_v.insert(0, SetPatchEvent(patch))

    def set_volume(self, volume):
        self.m_v.append(SetVolumeEvent(volume))

    def prepend_volume(self, volume):
        self.m_v.insert(0, SetVolumeEvent(volume))

    def set_bpm(self, bpm, notelen=4):
        self.m_v.append(TempoEvent(bpm, notelen))

    def prepend_bpm(self, bpm, notelen=4):
        assert isinstance(bpm, int)
        self.m_v.insert(0, TempoEvent(bpm, notelen))

    def bender(self, chn, value):
        "value >= 0"
        self.m_v.append([self.BENDER, chn, value])

    def merge_with(self, B):
        D = {}
        for track in [self, B]:
            pos = Rat(0, 1)
            for event in track.m_v:
                if isinstance(event, Delay):
                    pos = pos + event.m_duration
                else:
                    if pos not in D:
                        D[pos] = []
                    D[pos].append(event)
        kv = list(D.keys())
        kv.sort()
        self.m_v = []
        for x in range(len(kv) - 1):
            for event in D[kv[x]]:
                self.m_v.append(event)
            self.m_v.append(Delay(kv[x + 1] - kv[x]))
        for event in D[kv[-1]]:
            self.m_v.append(event)

    def replace_note(self, old, new):
        assert isinstance(old, int)
        assert 0 <= old < 128
        assert isinstance(new, int)
        assert 0 <= new < 128
        for event in self.m_v:
            if isinstance(event, (NoteOnEvent, NoteOffEvent)) \
                    and event.m_pitch == old:
                event.m_pitch = new

    def calculate_event_times(self):
        """
        Set the variable m_time on each Event. Well actually we don't set
        it on the Delay events because events of that type does not generate
        any events when generating music.
        """
        pos = Rat(0, 1)
        for e in self.m_v:
            if isinstance(e, Delay):
                pos += e.m_duration
            else:
                e.m_time = pos

    def __add__(self, other):
        ret = Track()
        ret.m_v.extend(copy.deepcopy(self.m_v))
        ret.m_v.extend(copy.deepcopy(other.m_v))
        return ret


class PercussionTrack(Track):

    def __init__(self):
        Track.__init__(self)



#!/usr/bin/python3
# GNU Solfege - free ear training software
# Copyright (C) 2009, 2011, 2016 Tom Cato Amundsen
# Licence is GPL, see file COPYING

"""
Generate sha1 hash value of a file after removing lines
starting with # and empty lines.

This used to preserve the statistics of lesson files when
doing changes to them that does not affect the statistics.

Usage:
 ./tools/hash-of-file.py filename
"""

import sys
sys.path.insert(0, ".")

if len(sys.argv) < 2 or sys.argv[1] in ('-h', '--help'):
   import __main__
   print(__main__.__doc__)
   sys.exit()
import textwrap
from solfege import i18n
i18n.setup(".")
from solfege import statistics

print()
print("\n".join(textwrap.wrap(
 "To preserve statistics, place the following in the header "
 "of %s before making changes to it:" % sys.argv[1])))
print()
print("    replaces = \"%s\"" % statistics.hash_of_lessonfile(str(sys.argv[1])))
print()



# coding: utf-8
# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
TODO:
    - cfg.get_list
    - cfg.set_list
    - watches outside ConfigUtils
>>> from solfege import cfg
>>> cfg.initialise(None, "default.config", "~/.cfg-test")
>>> cfg.set_string("section/stringvar", "value")
>>> cfg.get_string("section/stringvar")
'value'
>>> cfg.set_int("section/intvar", 14)
>>> cfg.get_int("section/intvar")
14
>>> cfg.set_float("section/floatvar", 3.141592)
>>> cfg.get_float("section/floatvar") == 3.141592
1
>>> cfg.set_bool("section/boolvar", 4)
>>> cfg.get_bool("section/boolvar")
1
>>> cfg.set_bool("section/boolvar", 0)
>>> cfg.get_bool("section/boolvar")
0
>>> cfg.set_list("section/listv", [1, 2, 3])
>>> cfg.get_list("section/listv")
[1, 2, 3]
>>> c = ConfigUtils("section")
>>> c.get_string("section/stringvar")
'value'
>>> c.get_int("section/intvar")
14
>>> c.get_bool("section/boolvar")
0
>>> c.get_float("section/floatvar") == 3.141592
1
>>> cfg.initialise(None, None, "~/.solfege-cfg-testing")
>>> c.get_string("ss/a")
''
>>> c.get_string("ss/a=hei")
'hei'
>>> c.get_int("ss/a=3")
3
>>> c.get_list("ss/a")
[]
>>> c.get_list("ss/a=[1, 2, 3]")
[1, 2, 3]
>>> comment_re.match("# basd") != None
True
>>> section_re.match("#[section]") == None
True
>>> section_re.match(" [section]") == None
True
>>> section_re.match("[section]") != None
True
>>> value_re.match("key=value") != None
True
>>> value_re.match("#key=value") == None
True
>>> value_re.match(" key=value") == None
True
"""

import codecs
import logging
import os
import re

from solfege import filesystem

# match both "section-name" and "user:dir/module"
section_re = re.compile("^\[([\w\:\/-_]*?)\]")
value_re = re.compile("^([\w-]*?)=(.*)")
comment_re = re.compile("#.*")

_blocked_watches = {}
_watches = {}
_watch_counter = {}
_user_filename = None
_app_defaults_filename = None
_system_filename = None
data = {}


def split(key):
    v = key.rpartition("/")
    return v[0], v[2]


def parse_file_into_dict(dictionary, filename):
    with codecs.open(filename, "r", "utf-8") as f:
        section = None
        for lineno, line in enumerate(f):
            line = line.strip()
            section_m = section_re.match(line)
            value_m = value_re.match(line)
            comment_m = comment_re.match(line)
            if comment_m:
                pass
            elif section_m:
                section = section_m.groups()[0]
                if section not in dictionary:
                    dictionary[section] = {}
            elif value_m:
                assert section
                dictionary[section][value_m.groups()[0]] = value_m.groups()[1]
            elif not line.strip():
                pass
            else:
                raise CfgParseException(filename, lineno)
        f.close()
        return dictionary


def sync():
    dump(data, _user_filename)


def _maybe_create_key(key):
    section, k = split(key)
    if section not in data:
        data[section] = {}


def iterate_sections():
    for n in list(data.keys()):
        yield n
########################
## set_XXXX functions ##
########################


def set_string(key, val):
    _maybe_create_key(key)
    section, k = split(key)
    if not isinstance(val, str):
        val = val.decode("utf-8")
    oldval = get_string(key)
    data[section][k] = val
    if key in _watches and (oldval != get_string(key)):
        if key in _blocked_watches and _blocked_watches[key] > 0:
            return
        for cb in list(_watches[key].values()):
            cb(key)


def set_int(key, val):
    assert isinstance(val, int)
    _maybe_create_key(key)
    section, k = split(key)
    oldval = get_string(key)
    data[section][k] = str(val)
    if key in _watches and (oldval != get_string(key)):
        if key in _blocked_watches and _blocked_watches[key] > 0:
            return
        for cb in list(_watches[key].values()):
            cb(key)


def set_float(key, val):
    assert isinstance(val, (float, int))
    _maybe_create_key(key)
    section, k = split(key)
    oldval = get_string(key)
    data[section][k] = str(val)
    if key in _watches and (oldval != get_string(key)):
        if key in _blocked_watches and _blocked_watches[key] > 0:
            return
        for cb in list(_watches[key].values()):
            cb(key)


def set_bool(key, val):
    _maybe_create_key(key)
    if val:
        set_string(key, "true")
    else:
        set_string(key, "false")


def set_list(key, val):
    set_string(key, str(val))

########################
## get_XXX functions ##
########################


def get_string(key):
    if len(key.split("=")) == 2:
        key, default = key.split("=")
    else:
        default = ""
    section, k = split(key)
    try:
        return data[section][k]
    except KeyError:
        return default


def get_int(key):
    if len(key.split("=")) == 2:
        key, default = key.split("=")
    else:
        default = 0
    section, k = split(key)
    try:
        # UGH win32 fix
        if not data[section][k]:
            return 0
        return int(float(data[section][k]))
    except KeyError:
        return int(default)


def get_float(key):
    if len(key.split("=")) == 2:
        key, default = key.split("=")
    else:
        default = 0
    section, k = split(key)
    try:
        return float(data[section][k])
    except KeyError:
        return float(default)


def get_list(key):
    if len(key.split("=")) == 2:
        key, default = key.split("=")
    else:
        default = []
    section, k = split(key)
    try:
        return eval(data[section][k])
    except KeyError:
        return default


def get_bool(key):
    if get_string(key) == 'true':
        return 1
    elif get_string(key) == 'false':
        return 0
    else:
        return 0

#####################################


def del_key(key):
    section, k = split(key)
    # FIXME watches
    if section not in data:
        return
    if k not in data[section]:
        return
    del data[section][k]
    if not data[section]:
        del data[section]


def del_section(section):
    # FIXME watches
    if section not in data:
        # Trying to delete a deleted section
        return
    del data[section]

######################################
## reading and writing the database ##
######################################


def dump(datadict, fn):
    f = codecs.open(fn, 'w', 'utf-8')
    for section in datadict:
        f.write("[%s]\n" % section)
        for name in datadict[section]:
            f.write("%s=%s\n" % (name, datadict[section][name]))
        f.write("\n")
    f.close()


def drop_user_config():
    """
    Reread the config data, but only read the systems defaults from
    /usr/share/solfege/default.config and /etc/solfege, not
    $HOME/.solfegerc
    """
    global data, _watches, _watch_counter, _blocked_watches
    data = {}
    _watches = {}
    _watch_counter = 0
    _blocked_watches = {}
    data = parse_file_into_dict(data, _app_defaults_filename)
    if _system_filename and os.path.isfile(_system_filename):
        data = parse_file_into_dict(data, _system_filename)


def reread_data():
    global data, _watches, _watch_counter, _blocked_watches
    _watches = {}
    _watch_counter = 0
    _blocked_watches = {}
    if _app_defaults_filename:
        data = parse_file_into_dict({}, _app_defaults_filename)
    if _system_filename:
        data = parse_file_into_dict(data, _system_filename)
    if _user_filename and os.path.isfile(_user_filename):
        data = parse_file_into_dict(data, _user_filename)


class CfgParseException(Exception):

    def __init__(self, filename, lineno):
        Exception.__init__(self)
        self.m_filename = filename
        self.m_lineno = lineno

    def __str__(self):
        return "CfgParseException: Failed to parse line %(lineno)i of the file '%(filename)s'" % {'lineno': self.m_lineno, 'filename': self.m_filename}


def initialise(app_defaults_filename, system_filename, user_filename):
    """
    app_defaults_filename: file must exist!
    system_filename: sys admin can override app_defaults_filename
    user_filename: the state of the app is stored here, also user config.
    """
    global data, _watches, _watch_counter, _blocked_watches
    global _app_defaults_filename, _system_filename, _user_filename
    _watches = {}
    _watch_counter = 0
    _blocked_watches = {}
    _app_defaults_filename = app_defaults_filename
    _system_filename = system_filename
    _user_filename = user_filename = filesystem.expanduser(user_filename)
    if app_defaults_filename:
        try:
            data = parse_file_into_dict({}, app_defaults_filename)
        except CfgParseException:
            print("-" * 60)
            print("The program failed to parse default.config. This means that")
            print("you have modified the file, because it was ok then Solfege")
            print("was released. Fix the file yourself, or reinstall Solfege.")
            print("-" * 60)
            raise
    else:
        data = {}
    if system_filename and os.path.isfile(system_filename):
        data = parse_file_into_dict(data, system_filename)
    if os.path.isfile(user_filename):
        data = parse_file_into_dict(data, user_filename)


class ConfigUtils(object):

    def __init__(self, exname):
        self.m_exname = exname

    def block_watch(self, name):
        """
        Stop functions watching a name being called. You have to call the
        same number of unblock_watch calls as block_calls to remove a blocking.
        """
        name = self._expand_name(name)
        if name not in _blocked_watches:
            _blocked_watches[name] = 0
        _blocked_watches[name] += 1

    def unblock_watch(self, name):
        name = self._expand_name(name)
        assert name in _blocked_watches
        assert _blocked_watches[name] > 0
        _blocked_watches[name] -= 1

    def add_watch(self, name, callback):
        global _watch_counter
        name = self._expand_name(name)
        if name not in _watches:
            _watches[name] = {}
        _watches[name][_watch_counter] = callback
        _watch_counter += 1
        return _watch_counter - 1

    def remove_watch(self, name, watch_id):
        name = self._expand_name(name)
        if name in _watches:
            if watch_id in _watches[name]:
                del _watches[name][watch_id]
                if _watches[name] == {}:
                    del _watches[name]
            else:
                logging.warning("cfg.remove_watch: id don't exist: %s", watch_id)
        else:
            logging.warning("cfg.remove_watch: name is not watched: %s", name)

    def _expand_name(self, name):
        if name.count("/") == 0:
            return "%s/%s" % (self.m_exname, name)
        return name
    ######
    # set
    ######

    def set_string(self, name, val):
        set_string(self._expand_name(name), val)

    def set_int(self, name, val):
        set_int(self._expand_name(name), val)

    def set_float(self, name, val):
        set_float(self._expand_name(name), val)

    def set_bool(self, name, val):
        set_bool(self._expand_name(name), val)

    def set_list(self, name, val):
        set_list(self._expand_name(name), val)
    ######
    # get
    ######

    def _get(self, func, name, default=""):
        return func(self._expand_name(name) + default)

    def get_string(self, name):
        r = self._get(get_string, name)
        if r is None:
            return ""
        else:
            return r

    def get_int(self, name):
        return self._get(get_int, name)

    def get_int_with_default(self, name, default):
        assert isinstance(default, int)
        return self.get_int(name + "=%i" % default)

    def get_float(self, name):
        return self._get(get_float, name)

    def get_bool(self, name):
        return self._get(get_bool, name)

    def get_list(self, name):
        try:
            if '=' in name:
                return eval(self._get(get_string, name))
            return eval(self._get(get_string, name, "=[]"))
        except SyntaxError:
            return []
